# pylint: disable=redefined-outer-name

import json
import os
from typing import List

import pytest

from flaat import BaseFlaat
from flaat.requirements import (
    AllOf,
    HasClaim,
    HasSubIss,
    IsTrue,
    N_Of,
    OneOf,
    Requirement,
    Satisfied,
    Unsatisfiable,
    get_claim_requirement,
    get_vo_requirement,
    get_audience_requirement,
)
from flaat.test_env import (
    FLAAT_ISS,
    User,
    AUD_OIDC_AGENT_ACCOUNT,
    AUD_FLAAT_ISS,
    load_at,
)
from flaat.exceptions import FlaatException

INVALID_ENTITLEMENT = "foo-bar"
VALID_ENTITLEMENT = "urn:mace:egi.eu:group:eosc-synergy.eu:role=member#aai.egi.eu"
CLAIM = "eduperson_entitlement"

INVALID_AUDIENCE = "foo"
VALID_AUDIENCE = "test-audience"


class RequirementsUser(User):
    def __init__(self):
        flaat = BaseFlaat()
        flaat.set_issuer(FLAAT_ISS)
        super().__init__(flaat)

        self.success_requirements: List[Requirement] = [
            HasSubIss(),
            Satisfied(),
            IsTrue(lambda _: True),
        ]
        for match in [1, "all"]:
            self.success_requirements.append(
                get_claim_requirement(self.groups, claim=self.claim_groups, match=match)
            )
            self.success_requirements.append(
                get_vo_requirement(
                    self.entitlements, claim=self.claim_entitlements, match=match
                )
            )

        self.success_requirements.append(AllOf(*self.success_requirements))
        self.success_requirements.append(N_Of(2, *self.success_requirements))
        self.success_requirements.append(OneOf(*self.success_requirements))

        self.failure_requirements = [Unsatisfiable()]


def _get_user_infos(flaat, access_token):
    user_infos = flaat.get_user_infos_from_access_token(access_token)
    if user_infos is None:
        raise FlaatException(
            "Cannot run tests: could not fetch a userinfo with the access token"
        )
    return user_infos


@pytest.fixture
def user():
    return RequirementsUser()


@pytest.fixture
def flaat_aud():
    if AUD_OIDC_AGENT_ACCOUNT == "" or AUD_FLAAT_ISS == "":
        pytest.skip("No env vars set for OP that supports audience")
    flaat = BaseFlaat()
    flaat.set_issuer(AUD_FLAAT_ISS)
    flaat.set_trusted_OP_list([AUD_FLAAT_ISS])
    return flaat


def test_possible_requirements_success(user):
    """constructs (nearly) all options of using our decorators for testing them
    against the frameworks"""

    for req in user.success_requirements:
        assert req.is_satisfied_by(user.user_infos).is_satisfied


def test_possible_requirements_failure(user):
    for req in user.failure_requirements:
        assert not req.is_satisfied_by(user.user_infos).is_satisfied


def test_empty_meta_requirements(user):
    user_infos = user.user_infos
    assert not AllOf().is_satisfied_by(user_infos).is_satisfied
    assert not OneOf().is_satisfied_by(user_infos).is_satisfied
    assert not N_Of(1).is_satisfied_by(user_infos).is_satisfied


def test_unset_or_unsupported_aud_success(user):
    user_infos = user.user_infos
    assert get_audience_requirement("").is_satisfied_by(user_infos).is_satisfied
    assert get_audience_requirement([]).is_satisfied_by(user_infos).is_satisfied


def test_supported_aud_missing(flaat_aud):
    access_token = load_at(AUD_OIDC_AGENT_ACCOUNT, min_valid_period=3600)
    user_infos = _get_user_infos(flaat_aud, access_token)
    assert (
        not get_audience_requirement(INVALID_AUDIENCE)
        .is_satisfied_by(user_infos)
        .is_satisfied
    )


def test_supported_aud_success(flaat_aud):
    access_token = load_at(
        AUD_OIDC_AGENT_ACCOUNT, min_valid_period=3600, audience=VALID_AUDIENCE
    )
    user_infos = _get_user_infos(flaat_aud, access_token)
    assert (
        get_audience_requirement(VALID_AUDIENCE)
        .is_satisfied_by(user_infos)
        .is_satisfied
    )
    assert (
        get_audience_requirement([VALID_AUDIENCE])
        .is_satisfied_by(user_infos)
        .is_satisfied
    )
    assert (
        get_audience_requirement([VALID_AUDIENCE, INVALID_AUDIENCE])
        .is_satisfied_by(user_infos)
        .is_satisfied
    )


def test_supported_aud_invalid(flaat_aud):
    access_token = load_at(
        AUD_OIDC_AGENT_ACCOUNT, min_valid_period=3600, audience=VALID_AUDIENCE
    )
    user_infos = _get_user_infos(flaat_aud, access_token)
    assert (
        not get_audience_requirement(INVALID_AUDIENCE)
        .is_satisfied_by(user_infos)
        .is_satisfied
    )


def test_supported_aud_multiple(flaat_aud):
    access_token = load_at(
        AUD_OIDC_AGENT_ACCOUNT, min_valid_period=3600, audience=f"{VALID_AUDIENCE} bar"
    )
    user_infos = _get_user_infos(flaat_aud, access_token)
    assert (
        get_audience_requirement(VALID_AUDIENCE)
        .is_satisfied_by(user_infos)
        .is_satisfied
    )
    assert (
        get_audience_requirement([VALID_AUDIENCE, INVALID_AUDIENCE])
        .is_satisfied_by(user_infos)
        .is_satisfied
    )
    assert (
        not get_audience_requirement(INVALID_AUDIENCE)
        .is_satisfied_by(user_infos)
        .is_satisfied
    )



"""
This module contains classes to express diverse requirements which a user needs to satisfy in order to use a view function.

The convenience functions :meth:`get_claim_requirement` and :meth:`get_vo_requirement` are recommended to construct individual requirements.

If you want to combine multiple requirements use the "meta requirements"  :class:`AllOf` and :class:`N_Of`.
"""
import logging
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union

import aarc_entitlement

from flaat.exceptions import FlaatException
from flaat.user_infos import UserInfos

logger = logging.getLogger(__name__)


# No leading slash ('/') in ops_that_support_audience !!!
OPS_THAT_SUPPORT_AUDIENCE = [
    "https://iam-test.indigo-datacloud.eu",
    "https://iam.deep-hybrid-datacloud.eu",
    "https://iam.extreme-datacloud.eu",
    "https://wlcg.cloud.cnaf.infn.it",
]


@dataclass
class CheckResult:
    """CheckResult is the result of an `is_satisfied_by` check"""

    is_satisfied: bool
    """ Is True if the requirement was satisfied by the user info """

    message: str
    """ Message describing the check result. This could be an error message. """

    data: Optional[Any] = None

    def render(self) -> Union[dict, str]:
        if self.data is None:
            return self.message

        return {
            "check": self.message,
            "check_details": self.data,
        }


class Requirement:
    """Requirement is the base class of all requirements.
    The have a method `is_satisfied_by` which returns a `CheckResult` instance.
    """

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        _ = user_infos
        return CheckResult(False, "method not overwritten")


# REQUIREMENT is the type of requirements, either lazy or not
REQUIREMENT = Union[Requirement, Callable[[], Requirement]]

REQUEST_REQUIREMENT = Callable[[UserInfos, tuple, dict], CheckResult]


class Satisfied(Requirement):
    """Satisfied is always satisfied"""

    def is_satisfied_by(self, _):
        return CheckResult(True, "Requirement is always satisfied")


class Unsatisfiable(Requirement):
    """Unsatisfiable is never satisfied"""

    def is_satisfied_by(self, _):
        return CheckResult(False, "Requirement is unsatisfiable")


class IsTrue(Requirement):
    """
    IsTrue is satisfied if the provided func evaluates to True

    :param func: A function that is used to determine if a user info
        satisfies custom requirements.
    """

    def __init__(self, func: Callable[[UserInfos], bool]):
        self.func = func

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        return CheckResult(
            self.func(user_infos), f"Evaluation of: {self.func.__name__}"
        )


class MetaRequirement(Requirement):
    """MetaRequirement is a requirements consisting of multiple sub-requirements
    Use the childs AllOf or N_Of directly.
    """

    def __init__(self, *reqs: REQUIREMENT):
        self._requirements: List[REQUIREMENT] = list(reqs)

    def add_requirement(self, req: REQUIREMENT):
        self._requirements.append(req)

    @property
    def requirements(self) -> List[Requirement]:
        """do the lazy loading of callables"""
        reqs = []
        for _req in self._requirements:
            req = _req() if callable(_req) else _req
            reqs.append(req)

        return reqs


class AllOf(MetaRequirement):
    """
    AllOf is satisfied if all of its sub-requirements are satisfied.
    If there are no sub-requirements, this class is never satisfied.
    """

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        if len(self.requirements) == 0:
            return CheckResult(False, "No sub-requirements")

        satisfied = True
        message = "All sub-requirements are satisfied"
        failed_checks = []

        for req in self.requirements:
            check_result = req.is_satisfied_by(user_infos)
            if not check_result.is_satisfied:
                failed_checks.append(check_result.render())
                satisfied = False

        if not satisfied:
            message = f"{self.__class__.__name__}: Unsatisfied sub-requirements"

        return CheckResult(satisfied, message, data=failed_checks)


class N_Of(MetaRequirement):
    """
    N_Of is satisfied if at least `n` of its sub-requirements are satisfied.
    If there are no sub-requirements, this class is never satisfied.
    """

    def __init__(self, n: int, *reqs: Requirement):
        super().__init__(*reqs)
        self.n = n

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        if len(self.requirements) == 0:
            return CheckResult(False, "No sub-requirements")

        failed_checks = []
        n = 0
        for req in self.requirements:
            check_result = req.is_satisfied_by(user_infos)
            if not check_result.is_satisfied:
                failed_checks.append(check_result.render())
            else:
                n += 1

        if n >= self.n:
            return CheckResult(True, f"{n} of {self.n} sub-requirments are satisfied")

        return CheckResult(
            False,
            f"Only {n} of {self.n} sub-requirments were satisfied",
            data=failed_checks,
        )


class OneOf(N_Of):
    """
    OneOf is satisfied if at least one of its sub-requirements are satisfied.
    If there are no sub-requirements, this class is never satisfied.
    """

    def __init__(self, *reqs: Requirement):
        super().__init__(1, *reqs)


def _match_to_meta_requirement(match: Union[str, int]) -> MetaRequirement:
    """translates a match argument to meta requirements
    Valid values are: "all", "one" or int"""

    logger.debug(f"meta requirement: match {match}")
    if match == "all":
        return AllOf()
    if match == "one":
        return N_Of(n=1)
    if isinstance(match, int):
        if match == 1:
            return N_Of(n=1)
        return N_Of(match)

    raise FlaatException(
        "Argument 'match' has invalid value: Must be 'all', 'one' or int"
    )


class HasSubIss(Requirement):
    """HasSubIss is satisfied if the user has a subject and an issuer"""

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        if user_infos is None:
            return CheckResult(False, "No valid user_infos found")

        if user_infos.subject != "" and user_infos.issuer != "":
            return CheckResult(
                True, "Valid user: {user_infos.subject} @ {user_infos.issuer}"
            )

        return CheckResult(False, "user_infos have no subject / issuer")


class HasClaim(Requirement):
    """HasClaim is satisfied if the user has the specified claim value"""

    def __init__(self, required, claim: str):
        """
        claim is the name of the claim.
        value is the value the claim needs to have
        """
        # try parsing the value, if it does not work revert to equal comparisons
        self.use_parse = True
        self.value = self.parse(required)
        if self.value is None:
            self.use_parse = False
            self.value = required

        self.claim = claim

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        value = user_infos.get(self.claim, None)
        if value is None:
            return CheckResult(False, f"Claim '{self.claim}' is not available")

        matched = False
        matched_value = None
        if isinstance(value, list):
            for val in value:
                if self.matches(self.value, self.parse(val)):
                    matched_value = val
                    matched = True
                    break
        else:
            if self.matches(self.value, self.parse(value)):
                matched_value = value
                matched = True

        if not matched:
            return CheckResult(
                False,
                f"User has no claim '{self.claim}' with value: '{self.value}'",
            )

        return CheckResult(
            True,
            f"Match for the required value '{self.value}' of claim '{self.claim}': '{matched_value}'",
        )

    def _parse(self, raw):
        """_parse can be overwritten by subclasses"""
        return raw

    def parse(self, raw):
        if self.use_parse:
            return self._parse(raw)
        return raw

    def _matches(self, required, available) -> bool:
        """_matches can be overwritten by subclasses"""
        return required == available

    def matches(self, required, available) -> bool:
        if self.use_parse:
            return self._matches(required, available)
        return required == available


class HasAudience(HasClaim):
    """HasAudience is satisfied if the user's access token was issued for a specific audience"""

    def is_satisfied_by(self, user_infos: UserInfos) -> CheckResult:
        if (
            user_infos is not None
            and user_infos.issuer.rstrip("/") not in OPS_THAT_SUPPORT_AUDIENCE
        ):
            logger.warning(
                "Issuer %s does not support audience setting, ignoring audience requirement.",
                user_infos.issuer,
            )
            return CheckResult(
                True,
                "Issuer does not support audience setting, ignoring audience requirement.",
            )
        return super().is_satisfied_by(user_infos)


class HasAARCEntitlement(HasClaim):
    """HasAARCEntitlement is satisfies if the user has the provided AARC-G002/G069 entitlement
    If the argument `required` is not a parseable AARC entitlement, we revert to equals comparisons.
    """

    def _parse(self, raw: str):
        try:
            return aarc_entitlement.G069(raw)
        except aarc_entitlement.Error as e:
            logger.debug("Error parsing aarc entitlement: %s", e)
            return None

    def _matches(
        self, required: aarc_entitlement.Base, available: aarc_entitlement.Base
    ) -> bool:
        return available is not None and available.satisfies(required)


def _get_claim_requirement(
    required: Union[str, List[str]],
    claim: str,  # claim in the user info
    match: Union[str, int] = "all",
    claim_requirement_class=HasClaim,
) -> Requirement:
    """
    :param claim_requirement_class: If the claim values need specific handling this can be used to specify a class
        for the handling, like in :meth:`get_vo_requirement`.
    """

    if isinstance(required, list):
        requirement = _match_to_meta_requirement(match)

        for req in required:
            requirement.add_requirement(claim_requirement_class(req, claim=claim))
    else:
        requirement = claim_requirement_class(required, claim=claim)

    return requirement


def get_claim_requirement(
    required: Union[str, List[str]],
    claim: str,  # claim in the user info
    match: Union[str, int] = "all",
) -> Requirement:
    """
    :param required: The claim values that the user needs to have.
    :param claim: The claim of the value in `required`, e.g. `eduperson_entitlement`.
    :param match: May be "all" if all required values need to be matched, or "one" or an integer if a specific amount needs to be matched.

    :return: A requirement that is satisfied if the user has the claim value(s) of `required`.
    """
    return _get_claim_requirement(required, claim, match, HasClaim)


def get_vo_requirement(
    required: Union[str, List[str]],
    claim: str,  # claim in the user info
    match: Union[str, int] = "all",
) -> Requirement:
    """Equivalent to :meth:`get_claim_requirement`, but works for both groups and AARC entitlements."""
    return _get_claim_requirement(
        required, claim, match, claim_requirement_class=HasAARCEntitlement
    )


def get_audience_requirement(required: Union[str, List[str]]) -> Requirement:
    """Equivalent to :meth:`get_claim_requirement`, but specific to audience claim."""
    if required == "" or required == []:
        logger.debug("Required audience empty or not specified.")
        return Satisfied()
    return _get_claim_requirement(
        required, "aud", "one", claim_requirement_class=HasAudience
    )



# Flaat example with FastAPI
import logging
from fastapi import Depends, FastAPI, Request, Response
from fastapi.security import HTTPBasicCredentials, HTTPBearer
from flaat import AuthWorkflow
from flaat.config import AccessLevel
from flaat.fastapi import Flaat
from flaat.requirements import CheckResult, HasSubIss, IsTrue
from flaat.requirements import get_claim_requirement
from flaat.requirements import get_vo_requirement


# ------------------------------------------------------------------
# Basic configuration example ---------------------------------------
logging.basicConfig(level="WARNING")
logging.getLogger("flaat").setLevel("DEBUG")
logging.getLogger("urllib3").setLevel("DEBUG")
ADMIN_EMAILS = ["admin@foo.org", "dev@foo.org"]


# Standard fastapi blueprint snippet, source:
# https://fastapi.tiangolo.com/advanced/security/oauth2-scopes/
app = FastAPI()
flaat = Flaat()
security = HTTPBearer()


# Set a list of access levels to use
def is_admin(user_infos):
    return user_infos.user_info["email"] in ADMIN_EMAILS


flaat.set_access_levels(
    [
        AccessLevel("user", HasSubIss()),
        AccessLevel("admin", IsTrue(is_admin)),
    ]
)

flaat.set_trusted_OP_list(
    [
        "https://aai-demo.egi.eu/oidc",
        "https://aai-demo.egi.eu/auth/realms/egi",
        "https://aai-dev.egi.eu/oidc",
        "https://aai.egi.eu/oidc/",
        "https://aai.egi.eu/auth/realms/egi",
        "https://accounts.google.com/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://b2access.eudat.eu/oauth2/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.elixir-czech.org/oidc/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://orcid.org/",
        "https://proxy.demo.eduteams.org",
        "https://services.humanbrainproject.eu/oidc/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://wlcg.cloud.cnaf.infn.it/",
        "https://proxy.eduteams.org/",
    ]
)


# ------------------------------------------------------------------
# Routes definition -------------------------------------------------
@app.get("/")
def root():
    text = """This is an example for using flaat with Flask.
    The following endpoints are available:
        /authenticated              Requires a valid user
        /authenticated_callback     Requires a valid user, uses a custom callback on error
        /authorized_level           Requires user to fit the specified access level
        /authorized_claim           Requires user to have one of two claims
        /authorized_vo              Requires user to have an entitlement
        /full_custom                Fully custom auth handling
    """
    return Response(text, media_type="text/plain")


# -------------------------------------------------------------------
# Endpoint which requires of an authenticated user ------------------
@app.get("/authenticated")
@flaat.is_authenticated()
def authenticated(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: there was a valid login"


# -------------------------------------------------------------------
# Instead of giving an error this will return the custom error
# response from `my_own_failure` -------------------------------------
def my_own_failure(exception, user_infos=None):
    return "Custom callback 'my_own_failure' invoked"


@app.get("/authenticated_callback")
@flaat.is_authenticated(on_failure=my_own_failure)
def authenticated_callback(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: there was a valid login"


# -------------------------------------------------------------------
# Endpoint which requires an access level ---------------------------
@app.get("/authorized_level")
@flaat.access_level("admin")
def authorized_level(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: user has the required rights"


# -------------------------------------------------------------------
# The user needs to satisfy a certain requirement -------------------
email_requirement = get_claim_requirement(
    ["admin@foo.org", "dev@foo.org"],
    claim="email",
    match=1,
)

@app.get("/authorized_claim")
@flaat.requires(email_requirement)
def authorized_claim(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: User has the claim"


# -------------------------------------------------------------------
# The user needs belong to a certain virtual organization -----------
vo_requirement = get_vo_requirement(
    [
        "urn:mace:egi.eu:group:eosc-synergy.eu",
        "urn:mace:egi.eu:group:mteam.data.kit.edu:perfmon.m.d.k.e",
    ],
    "eduperson_entitlement",
    match=2,
)

@app.get("/authorized_vo")
@flaat.requires(vo_requirement)
def authorized_vo(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: user has the required entitlement(s)"


# -------------------------------------------------------------------
# For maximum customization use AuthWorkflow ------------------------
def my_request_check(user_infos, *args, **kwargs):
    if len(args) != 1:
        return CheckResult(False, "Missing request object")
    return CheckResult(True, "The request is allowed")


def my_process_args(user_infos, *args, **kwargs):
    """We can manipulate the view functions arguments here The user is
    already authenticated at this point, therefore we have `user_infos`,
    therefore we can base our manipulations on the users identity.
    """
    kwargs["email"] = user_infos.get("email", "")
    return (args, kwargs)


custom = AuthWorkflow(
    flaat,  # needs the flaat instance
    user_requirements=get_claim_requirement("bar", "foo"),
    request_requirements=my_request_check,
    process_arguments=my_process_args,
    on_failure=my_own_failure,
    ignore_no_authn=True,  # Don't fail if there is no authentication
)


@app.get("/full_custom")
@custom.decorate_view_func  # invoke the workflow here
def full_custom(
    request: Request,
    credentials: HTTPBasicCredentials = Depends(security),
):
    user_infos = flaat.get_user_infos_from_request(request)
    return "This worked: user has the required entitlement"


# -------------------------------------------------------------------
# Main function -----------------------------------------------------
if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8081)



# pylint: disable=redefined-outer-name

import logging

import pytest
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient

from flaat.fastapi import Flaat
from flaat.test_env import FLAAT_TRUSTED_OPS_LIST, User, get_status_kwargs_list

logger = logging.getLogger(__name__)

flaat = Flaat()
flaat.set_trusted_OP_list(FLAAT_TRUSTED_OPS_LIST)

DECORATORS = User(flaat).get_named_decorators()


async def view_func(request: Request, test_inject=None):
    _ = request
    _ = test_inject
    return {"message": "Success"}


@pytest.fixture
def app():
    """fastapi app testing fixture"""
    app = FastAPI()
    for decorator in DECORATORS:
        app.get(f"/{decorator.name}")(decorator.decorator(view_func))
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.mark.parametrize("status,kwargs", get_status_kwargs_list())
@pytest.mark.parametrize("decorator", DECORATORS)
def test_decorator(client, decorator, status, kwargs):
    resp = client.get(f"/{decorator.name}", **kwargs)
    expected = decorator.get_expected_status_code(status)
    assert resp.status_code == expected



from typing import Any, Optional


class FlaatException(Exception):
    """
    An error occured inside flaat. The cause can be misconfiguration
    or other not user related errors.
    This exception will cause a response with status 500 if unhandled.
    """

    name = "Error"
    status_code = 500
    data: Optional[Any] = None

    def render(self) -> dict:
        data = {
            "error": self.name,
            "error_description": str(self),
        }
        if self.data is not None:
            data["error_details"] = self.data
        return data


class FlaatForbidden(FlaatException):
    """
    The user is forbidden from using the service.
    This exception will cause a response with status 403 if unhandled.
    """

    name = "Forbidden"
    status_code = 403


class FlaatUnauthenticated(FlaatException):
    """
    The users identity could not be determined. Probably there was no access token
    or the access tokens issuer could not be determined.

    This exception will cause a response with status 401 if unhandled.
    """

    name = "Unauthenticated"
    status_code = 401



import logging
from dataclasses import dataclass
from typing import List, Union

from flaat.requirements import REQUIREMENT, HasSubIss, Satisfied, Unsatisfiable

logger = logging.getLogger(__name__)

LOG_LEVEL_OVERRIDES = {
    "requests_cache": logging.WARN,
    "urllib3": logging.WARN,
    "asyncio": logging.WARN,
}


def _apply_log_level_overrides():
    for name, level in LOG_LEVEL_OVERRIDES.items():
        logging.getLogger(name).setLevel(level)


@dataclass
class AccessLevel:
    """Access levels are basically named requirements.
    An example would be two access levels 'user' and 'admin', which have requirements for the respective level.

    In order to use an access level with a flaat instance, you need to use :meth:`flaat.BaseFlaat.set_access_levels` to add them to the list.
    """

    name: str
    """ The name of the access level. This is used in  :meth:`flaat.BaseFlaat.access_level` to identify this access level."""

    requirement: REQUIREMENT
    """ The requirement that users of this access level need to satisfy.
    If this is a callable, then the requirement is lazily loaded at runtime using the callable.
    """


# DEFAULT VALUES
Anyone = AccessLevel("ANYONE", Satisfied())
NoOne = AccessLevel("NOONE", Unsatisfiable())
Identified = AccessLevel("IDENTIFIED", HasSubIss())

DEFAULT_ACCESS_LEVELS = [
    Anyone,
    NoOne,
    Identified,
]

# No leading slash ('/') in ops_that_support_jwt !!!
OPS_THAT_SUPPORT_JWT = [
    "https://aai-dev.egi.eu/oidc",
    "https://aai.egi.eu/oidc",
    "https://aai-dev.egi.eu/auth/realms/egi",
    "https://aai-demo.egi.eu/auth/realms/egi",
    "https://aai.egi.eu/auth/realms/egi",
    "https://aai-dev.egi.eu/auth/realms/egi",
    "https://b2access-integration.fz-juelich.de/oauth2",
    "https://b2access.eudat.eu/oauth2",
    "https://iam-test.indigo-datacloud.eu",
    "https://iam.deep-hybrid-datacloud.eu",
    "https://iam.extreme-datacloud.eu",
    "https://login-dev.helmholtz.de/oauth2",
    "https://login.elixir-czech.org/oidc",
    "https://login.helmholtz-data-federation.de/oauth2",
    "https://login.helmholtz.de/oauth2",
    "https://oidc.scc.kit.edu/auth/realms/kit",
    "https://services.humanbrainproject.eu/oidc",
    "https://unity.helmholtz-data-federation.de/oauth2",
    "https://wlcg.cloud.cnaf.infn.it",
]


class FlaatConfig:
    """
    The configuration for Flaat instances.
    """

    def __init__(self):
        self.trusted_op_list: List[str] = []
        self.iss: str = ""
        self.client_id: str = ""
        self.client_secret: str = ""
        self.request_timeout: float = 1.2  # seconds
        self.verify_tls = True
        self.verify_jwt = True

        # access levels for the self.access_level decorator
        self.access_levels = DEFAULT_ACCESS_LEVELS

    def set_access_levels(self, access_levels: List[AccessLevel]):
        """
        Set the list of access levels for use with :meth:`flaat.BaseFlaat.access_level`.
        This list will overwrite the default access levels.

        :param access_level: List of :class:`AccessLevel` instances.
        """
        self.access_levels = access_levels

    def set_verbosity(self, verbosity: int, set_global=False):
        """
        Set the logging verbosity.

        :param verbosity: Verbosity integer from 0 (= error) to 3 (= debug)
        :param set_global: If set to `True` the logging level will be set for all loggers
        """
        if verbosity < 0 or verbosity > 3:
            raise ValueError("Verbosity needs to be [0-3]")

        level = {
            0: logging.ERROR,
            1: logging.WARN,
            2: logging.INFO,
            3: logging.DEBUG,
        }[verbosity]

        if set_global:
            logging.getLogger().setLevel(level)
            logger.debug("Setting global log level to: %s", level)
        else:
            logger.setLevel(level)
            logger.debug("Setting flaat log level to: %s", level)

        _apply_log_level_overrides()

    def set_issuer(self, issuer: str):
        """Pins the given issuer. Only users of this issuer will be able to use services.

        :param issuer: Issuer URL of the pinned issuer.
        """
        self.iss = issuer.rstrip("/")

    def set_trusted_OP_list(self, trusted_op_list: List[str]):
        """
        Sets a list of OIDC providers that you trust. This means that users of these OPs will be able
        to use your services.

        :param trusted_op_list: A list of the issuer URLs that you trust.
            An example issuer is: 'https://iam.deep-hybrid-datacloud.eu/'.
        """

        self.trusted_op_list = list(map(lambda iss: iss.rstrip("/"), trusted_op_list))

    def set_verify_tls(self, verify_tls=True):
        """*Only* use for development and debugging.
        Set to `False` to skip TLS certificate verification while processing requests.
        """
        self.verify_tls = verify_tls

    def set_verify_jwt(self, verify_jwt=True):
        """Set to `False` to skip JWT verification while processing requests."""
        self.verify_jwt = verify_jwt

    def set_client_id(self, client_id=""):
        """Set a client id for token introspection"""
        # FIXME: consider client_id/client_secret per OP.
        self.client_id = client_id

    def set_client_secret(self, client_secret=""):
        """Set a client secret for token introspection"""
        self.client_secret = client_secret

    def set_request_timeout(self, timeout: float = 1.2):
        """
        Set the timeout for individual requests (retrieving issuer configs, user infos and introspection infos).
        Note that the total runtime of a decorator could be significantly more, based on your `trusted_op_list`.

        :param timeout: Request timeout in seconds.
        """
        self.request_timeout = timeout



""" These tests are nothing AIO specific, but we can only test them with a framework.
We therefore picked AIO as an example.

FIXME The environment overrides in here are crude and will not work if run concurrently!
"""
import os
from typing import Optional

import pytest
from aiohttp import web

from flaat import ENV_VAR_AUTHN_OVERRIDE, ENV_VAR_AUTHZ_OVERRIDE, AuthWorkflow
from flaat.aio import Flaat
from flaat.aio.aio_test import flaat
from flaat.requirements import get_claim_requirement
from flaat.user_infos import UserInfos


@pytest.fixture
def app():
    """aio web Application for override testing"""

    workflow = AuthWorkflow(
        flaat,
        user_requirements=get_claim_requirement("claim-value", "claim-name"),
    )

    async def view_func(request, test_inject=None):
        _ = request
        _ = test_inject
        return web.Response(text="Success")

    _app = web.Application()
    _app.router.add_get("/", workflow.decorate_view_func(view_func))
    return _app


# from: https://docs.aiohttp.org/en/stable/testing.html#pytest-example
@pytest.fixture
def cli(event_loop, aiohttp_client, app):
    return event_loop.run_until_complete(aiohttp_client(app))


def _mock_auth_user(monkeypatch, user_infos: Optional[UserInfos]):
    def _mock_authenticate_user(*_, **__):
        return user_infos

    monkeypatch.setattr(Flaat, "authenticate_user", _mock_authenticate_user)


async def test_env_override_authentication(monkeypatch, cli):
    _mock_auth_user(monkeypatch, user_infos=None)

    resp = await cli.get("/")
    assert resp.status == 401  # no user

    os.environ[ENV_VAR_AUTHN_OVERRIDE] = "YES"

    resp = await cli.get("/")
    assert resp.status == 200  # no user, but authentication override

    del os.environ[ENV_VAR_AUTHN_OVERRIDE]


async def test_env_override_authorization(monkeypatch, cli):
    # this user does not meet the requirments from the workflow in this module
    forbidden_user = UserInfos(
        None,
        {"sub": "foo", "iss": "bar"},  # userinfo
        None,
    )
    _mock_auth_user(monkeypatch, user_infos=forbidden_user)

    resp = await cli.get("/")
    assert resp.status == 403  # user, but forbidden

    os.environ[ENV_VAR_AUTHN_OVERRIDE] = "YES"

    resp = await cli.get("/")
    assert resp.status == 200  # user, but authorization override

    del os.environ[ENV_VAR_AUTHN_OVERRIDE]


async def test_env_override_authorization_without_user(monkeypatch, cli):
    _mock_auth_user(monkeypatch, user_infos=None)

    resp = await cli.get("/")
    assert resp.status == 401  # no user, no override

    os.environ[ENV_VAR_AUTHZ_OVERRIDE] = "YES"

    resp = await cli.get("/")
    assert resp.status == 401  # no user, only authorization override

    del os.environ[ENV_VAR_AUTHZ_OVERRIDE]



# This code is distributed under the MIT License

import base64
from dataclasses import dataclass
import logging
from typing import Optional, List

import jwt

from flaat.exceptions import FlaatUnauthenticated
from flaat.issuers import IssuerConfig

logger = logging.getLogger(__name__)

# Expand this list in a sensible way
PERMITTED_SIGNATURE_ALGORITHMS = [
    "RS256",
    "RS384",
    "RS512",
    "EdDSA",
    "ES256",
    "ES256K",
    "ES384",
    "ES512",
    "ES521",
    "HS256",
    "HS384",
    "HS512",
    "PS256",
    "PS384",
    "PS512",
]


def _base64_url_encode(data):
    """Decode base64 encode data"""
    if not isinstance(data, bytes):
        data = data.encode("utf-8")
    encode = base64.urlsafe_b64encode(data)
    return encode.decode("utf-8").rstrip("=")


@dataclass
class AccessTokenInfo:
    """Infos from a JWT access token"""

    header: dict
    """ The JWTs JOSE header """

    body: dict
    """ The JWTs data payload """

    signature: str
    """ The JWTs JWS signature """

    verification: Optional[dict]
    """ Infos about the verification of the JWT.
    If set to `None`, then the JWT data is unverified."""

    def __init__(self, complete_decode, verification):
        self.header = complete_decode.get("header", {})
        self.body = complete_decode.get("payload", {})
        self.signature = _base64_url_encode(complete_decode.get("signature", b""))
        self.verification = verification

    @property
    def issuer(self) -> str:
        return self.body.get("iss", "")


class FlaatPyJWKClient(jwt.PyJWKClient):
    """Fixes the jwt.PyJWKClient class:

    * get_signing_keys
        * does not call self.get_jwk_set(), since it fails when "enc" keys are present
        * returns only keys used for signing (e.g. filters out keys with "use" == "enc")
    * get_signing_key_from_jwt
        * tries to retrieve keys by id only if "kid" is specified in token header
        * otherwise, it tries to infer the key type ("kty") from the algorithm used to sign the token ("alg")
        * "alg" is always present in JWT header
    * an additional method get_signing_key_by_alg
    """

    def get_signing_keys(self) -> List[jwt.api_jwk.PyJWK]:
        data = self.fetch_data()
        # filter for signing keys, i.e. "use" in ["sig", None]
        keys = [
            key for key in data.get("keys", []) if key.get("use", None) in ["sig", None]
        ]
        signing_keys = jwt.PyJWKSet(keys)
        if not signing_keys:
            raise jwt.exceptions.PyJWKClientError(
                "The JWKS endpoint did not contain any signing keys"
            )

        return signing_keys.keys

    def get_signing_key_by_alg(self, alg: str) -> jwt.api_jwk.PyJWK:
        # algorithm is none, then signing key is None; signature must be empty octet string
        if alg == "none":
            return jwt.api_jwk.PyJWK({}, algorithm="none")
        # infer key type from algorithm
        key_type = ""
        if alg.startswith("RS") or alg.startswith("PS"):
            key_type = "RSA"
        if alg.startswith("HS"):
            key_type = "oct"
        if alg.startswith("ES"):
            key_type = "EC"
        if alg.startswith("Ed"):
            key_type = "OKP"

        signing_keys = self.get_signing_keys()
        signing_key = None

        for key in signing_keys:
            if key.key_type == key_type:
                signing_key = key
                break

        if not signing_key:
            raise jwt.exceptions.PyJWKClientError(
                f'Unable to find a signing key that matches alg: "{alg}"'
            )

        return signing_key

    def get_signing_key_from_jwt(self, token: str) -> jwt.api_jwk.PyJWK:
        unverified = jwt.api_jwt.decode_complete(
            token, options={"verify_signature": False}
        )
        header = unverified["header"]

        kid = header.get("kid", None)
        if kid:
            return self.get_signing_key(kid)

        # alg MUST be present, possible values defined at https://datatracker.ietf.org/doc/html/rfc7518#section-3.1
        alg = header.get("alg", None)
        if alg:
            return self.get_signing_key_by_alg(alg)

        raise FlaatUnauthenticated(
            "Could not verify JWT: The token header did not contain an 'alg'."
        )


def get_access_token_info(access_token, verify=True) -> Optional[AccessTokenInfo]:
    unverified = {}
    try:
        unverified = jwt.api_jwt.decode_complete(
            access_token,
            options={
                "verify_signature": False,
            },
        )
        unverified_body = unverified.get("payload", {})
    except jwt.DecodeError:
        return None

    if not verify:
        return AccessTokenInfo(unverified, verification=None)

    issuer = IssuerConfig.get_from_string(unverified_body.get("iss", ""))
    if issuer is None:
        raise FlaatUnauthenticated("Could not verify JWT: No 'iss' claim in body")

    jwks_uri = issuer.issuer_config.get("jwks_uri", "")
    if jwks_uri == "":
        raise FlaatUnauthenticated(
            "Could not verify JWT: Issuer config has no jwks_uri"
        )

    jwk_client = FlaatPyJWKClient(jwks_uri)
    signing_key = jwk_client.get_signing_key_from_jwt(access_token)

    try:
        complete_decode = jwt.api_jwt.decode_complete(
            access_token,
            signing_key.key,
            algorithms=PERMITTED_SIGNATURE_ALGORITHMS,
            options={"verify_aud": False},
        )
    except jwt.exceptions.PyJWTError as e:
        raise FlaatUnauthenticated(f"Could not verify JWT: {e}") from e

    return AccessTokenInfo(
        complete_decode,
        verification={"algorithm": complete_decode.get("header", {}).get("alg", "")},
    )



import pytest

from flaat.issuers import IssuerConfig, is_url
from flaat.test_env import FLAAT_AT, FLAAT_ISS, environment


class TestURLs:
    def test_url_1(self):
        assert is_url("http://heise.de")

    def test_valid_url_http(self):
        assert is_url("http://heise.de")

    def test_valid_url_https(self):
        assert is_url("http://heise.de")

    def test_valid_url_ftp(self):
        assert is_url("http://heise.de")

    def test_valid_url_https_path(self):
        assert is_url("https://heise.de/thi_s&is=difficult")

    def test_invalid_url(self):
        assert not is_url("htp://heise.de")


def test_token_introspection():
    client_id = environment.get("FLAAT_CLIENT_ID")
    client_secret = environment.get("FLAAT_CLIENT_SECRET")
    if client_id is None or client_secret is None:  # pragma: no cover
        pytest.skip("FLAAT_CLIENT_ID and FLAAT_CLIENT_SECRET are not set")

    issuer_config = IssuerConfig.get_from_string(FLAAT_ISS)
    assert issuer_config is not None
    issuer_config.client_id = client_id
    issuer_config.client_secret = client_secret
    introspection_info = issuer_config._get_introspected_token_info(FLAAT_AT)
    assert introspection_info is not None



from json import JSONEncoder
import logging
from time import time
from typing import Any, Optional


logger = logging.getLogger(__name__)


class UserInfos:
    """Infos about an access token and the user it belongs to.
    This class acts like a dictionary with regards to claims.
    So `infos["foo"]` will give you the claim if it does exist in one of the underlying dicts.
    """

    user_info: dict
    """ user_info is the user info dictionary from the user info endpoint of the issuer."""
    access_token_info: Optional[Any] = None  # Optional AccessTokenInfo
    """ Is set to `AccessTokenInfo` if the respective access token was a JWT."""
    introspection_info: Optional[dict] = None
    """ Is the data returned from the token intropsection endpoint of the issuer."""

    def __init__(
        self,
        access_token_info: Optional[Any],  # Optional AccessTokenInfo
        user_info: dict,
        introspection_info: Optional[dict],
    ):
        self.access_token_info = access_token_info
        self.user_info = user_info
        self.introspection_info = introspection_info

        # trigger possible post processing here
        self.post_process_dictionaries()

    def _strip_duplicate_infos(self):  # pragma: no cover
        """strip duplicate infos from the introspection_info and access_token_info.body"""
        if self.introspection_info is not None:
            for key in self.user_info.keys():
                if key in self.introspection_info:
                    del self.introspection_info[key]
        if self.access_token_info is not None:
            for key in self.user_info.keys():
                if key in self.access_token_info.body:
                    del self.access_token_info.body[key]

    def post_process_dictionaries(self):
        """post_process_dictionaries can be used to do post processing on the raw dictionaries after initialization.
        Extend this class and overwrite this method to do custom post processing.
        Make sure to call `super().post_process_dictionaries()`, so the post processing done here is picked up.
        """
        # copy a possible 'iss' fields into the user info if it does not exist
        # This is useful if someone extracts only the user_info dictionary from us
        if "iss" not in self.user_info and self.has_key("iss"):
            self.user_info["iss"] = self["iss"]

        # striping duplicates is somewhat opinionated and is therefore not included here
        # self._strip_duplicate_infos()

    @property
    def valid_for_secs(self) -> Optional[int]:
        """Is set if we now about the expiry of these user infos."""

        def _timeleft(info_dict, claim="exp") -> Optional[int]:
            """Get expiry from info dictionary (either from access token or introspection"""
            if claim in info_dict:
                now = time()
                timeleft = info_dict[claim] - now
                return timeleft

            return None  # pragma: no cover

        timeleft = None
        if self.introspection_info is not None:
            timeleft = _timeleft(self.introspection_info)

        if timeleft is None and self.access_token_info is not None:
            timeleft = _timeleft(self.access_token_info.body)

        return timeleft

    @property
    def issuer(self) -> str:
        """The issuer of the access token"""
        return str(self.get("iss", ""))

    @property
    def subject(self) -> str:
        """The users subject at the issuer"""
        return str(self.get("sub", ""))

    # make the UserInfos act like a dictionary with regard to claims
    def __getitem__(self, key):
        if key in self.user_info:
            return self.user_info[key]
        if self.introspection_info is not None and key in self.introspection_info:
            return self.introspection_info[key]
        if self.access_token_info is not None and key in self.access_token_info.body:
            return self.access_token_info.body[key]
        raise KeyError(
            "Claim does not exist in user_info, access_token_info.body and introspection_info"
        )  # pragma: no cover

    def has_key(self, key):
        return (
            key in self.user_info
            or (
                self.access_token_info is not None
                and key in self.access_token_info.body
            )
            or (self.introspection_info is not None and key in self.introspection_info)
        )

    def get(self, key, default=None):
        if self.has_key(key):
            return self[key]
        return default

    def __str__(self):
        return f"{self.subject}@{self.issuer}"

    def toJSON(self) -> str:
        """Render these infos to JSON"""

        class ATEncoder(JSONEncoder):
            def default(self, o):
                return o.__dict__

        return ATEncoder(indent=4, sort_keys=True, separators=(",", ":")).encode(self)



import contextlib
import json
from unittest import mock

import pytest

import jwt
from flaat.access_tokens import get_access_token_info, FlaatPyJWKClient
from flaat.test_env import FLAAT_AT, NON_JWT_FLAAT_AT


class TestTokens:
    def test_get_accesstoken_info_non_jwt(self):
        access_token = NON_JWT_FLAAT_AT
        if access_token == "":
            pytest.skip("No non JWT access token")

        access_token_info = get_access_token_info(access_token)
        assert access_token_info is None

    def test_get_accesstoken_info_jwt(self):
        access_token = FLAAT_AT
        if access_token in ["", "mock_jwt_at"]:
            pytest.skip("No JWT access token")
        access_token_info = get_access_token_info(access_token)
        assert access_token_info is not None
        assert access_token_info.verification is not None


RESPONSE_DATA = {
    "keys": [
        {
            "kty": "RSA",
            "kid": "bilbo.baggins@hobbiton.example",
            "use": "sig",
            "n": "n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqVwGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuCLqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5gHdrNP5zw",
            "e": "AQAB",
        },
        {
            "kid": "O_YttVFkHHz5miLSoyU9s88nH7NZ3IjnXJdtdWZWET8",
            "kty": "RSA",
            "alg": "RSA-OAEP",
            "use": "enc",
            "n": "woLHcpbl4iV-E8UeXNE1Ne2iKwbEs653WE808JB5saqYIpwI9ItVvdP7dWb0Nx2EyO6YfwcfQ50O2wz5rQDkCiGNY9vF4wq3UV1VTYc-Ed8zlLC9hfJQXO0dNXyrdem4im0jb-4y9S2VGRGLTy_1P-MnN60Tqr9TVpR0XEx0sAS12xO9cp0JesfwIgmRdAi8D1A7ug1w35SIsnT5rzEqciu1o7GytcL7EuTy4MRNJp1nXN0fZuNYoIqq7qVbHNoi9WhC1yuPA6AoJnt11wjifJAWraRnMfTCfaeJ-tW43xjpZlOqAxaB-3r-UTj2JBGbIDvF-yeW5E2yjRD9YsWYDQ",
            "e": "AQAB",
            "x5c": [
                "MIIClTCCAX0CBgF++OQ2PzANBgkqhkiG9w0BAQsFADAOMQwwCgYDVQQDDANlZ2kwHhcNMjIwMjE0MTUzODMyWhcNMzIwMjE0MTU0MDEyWjAOMQwwCgYDVQQDDANlZ2kwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDCgsdyluXiJX4TxR5c0TU17aIrBsSzrndYTzTwkHmxqpginAj0i1W90/t1ZvQ3HYTI7ph/Bx9DnQ7bDPmtAOQKIY1j28XjCrdRXVVNhz4R3zOUsL2F8lBc7R01fKt16biKbSNv7jL1LZUZEYtPL/U/4yc3rROqv1NWlHRcTHSwBLXbE71ynQl6x/AiCZF0CLwPUDu6DXDflIiydPmvMSpyK7WjsbK1wvsS5PLgxE0mnWdc3R9m41igiqrupVsc2iL1aELXK48DoCgme3XXCOJ8kBatpGcx9MJ9p4n61bjfGOlmU6oDFoH7ev5ROPYkEZsgO8X7J5bkTbKNEP1ixZgNAgMBAAEwDQYJKoZIhvcNAQELBQADggEBAGzA1gBJwyQCjDAFjXCTtEhjJa7xXkNqtBchQlTIDpxDqrKsXLKsLLZzTdx1Bw6p6dLqrRoB6mBbcQz+zHYS6r6E1Vp3JcUFs3aZ9dtW1dowzGKv0JU+42c3Gwl3en5dr3yVhahLRLI0PJf8eFhz4fItGISTDM4ppPK/b2p1Aj03zezD7aKQ9jygnhsGQmHtaW6m8mMQ9a7YwqURd3+VlexbbvvwX6pbXIQj3DK15f2rSLjvdva9smy898ewNonfBpLPvPc5oKLWRf1OyWzpbdwaMXHnItpASpWLFEjb56MjzhAD9dZy4Vruzd45rwohhLXr+qnTqmD81gee9gl13oU="
            ],
            "x5t": "OaLNe4i3IrUkY7G_RhP_czaw1R0",
            "x5t#S256": "5qRFpp9YT000QmD8xuXFWExHTa3hG_jAXBqCYlw6Vj0",
        },
    ]
}


@contextlib.contextmanager
def mocked_response(data):
    with mock.patch("urllib.request.urlopen") as urlopen_mock:
        response = mock.Mock()
        response.__enter__ = mock.Mock(return_value=response)
        response.__exit__ = mock.Mock()
        response.read.side_effect = [json.dumps(data)]
        urlopen_mock.return_value = response
        yield urlopen_mock


MOCK_URL = "http://mock_url"


class TestFlaatPyJWKClient:
    """Use RSA key from PyJWKClient tests,
    Only test new/overwritten methods and add new test cases.
    """

    def test_get_signing_keys(self):
        with mocked_response(RESPONSE_DATA):
            jwks_client = FlaatPyJWKClient(MOCK_URL)
            signing_keys = jwks_client.get_signing_keys()

        assert len(signing_keys) == 1
        assert isinstance(signing_keys[0], jwt.api_jwk.PyJWK)

    def test_get_signing_key_by_alg(self):
        with mocked_response(RESPONSE_DATA):
            jwks_client = FlaatPyJWKClient(MOCK_URL)
            signing_key = jwks_client.get_signing_key_by_alg("RS256")

        assert isinstance(signing_key, jwt.api_jwk.PyJWK)
        assert signing_key.key_type == "RSA"

    def test_get_signing_key_from_jwt_by_kid(self):
        token = "eyJhbGciOiJSUzI1NiIsImN0eSI6IkpXVCIsImtpZCI6ImJpbGJvLmJhZ2dpbnNAaG9iYml0b24uZXhhbXBsZSJ9.eyJpc3MiOiJodHRwczovL2Rldi04N2V2eDlydS5hdXRoMC5jb20vIiwic3ViIjoiYVc0Q2NhNzl4UmVMV1V6MGFFMkg2a0QwTzNjWEJWdENAY2xpZW50cyIsImF1ZCI6Imh0dHBzOi8vZXhwZW5zZXMtYXBpIiwiaWF0IjoxNTcyMDA2OTU0LCJleHAiOjE1NzIwMDY5NjQsImF6cCI6ImFXNENjYTc5eFJlTFdVejBhRTJINmtEME8zY1hCVnRDIiwiZ3R5IjoiY2xpZW50LWNyZWRlbnRpYWxzIn0.FY-57Y3K04hjK53P6t7XXnM_PLIYZbV0b596UOzmIBWkAznzga6Rqo-_uommL2hcsZMrzUtNpN0b9_11z7DDjaoPUYoJITyFDgJGLynMC538iLBWA-7x-3y-oKZkAK78yM5h5C3lIiRAlPKP_2UNyK-W40xyxoBW975fLqBVMChDUmQkyhH2GS4i16nZdbCYVMjGytxTHGH6810QneKVeoV0EStjxHjBKxTF26_1PRqeuMiYom6CRp7BdGQidDO_JxH7BqD6GPwnV3AzaFBnFsE5L9mrSTOymuvCELXLJwQYYGpT5i1ti4MP2jtSQYxvy3Zel56ybnSaaI1QTyRNAQ"

        with mocked_response(RESPONSE_DATA):
            jwks_client = FlaatPyJWKClient(MOCK_URL)
            signing_key = jwks_client.get_signing_key_from_jwt(token)

        data = jwt.decode(
            token,
            signing_key.key,
            algorithms=["RS256"],
            audience="https://expenses-api",
            options={"verify_exp": False},
        )

        assert data == {
            "iss": "https://dev-87evx9ru.auth0.com/",
            "sub": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC@clients",
            "aud": "https://expenses-api",
            "iat": 1572006954,
            "exp": 1572006964,
            "azp": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC",
            "gty": "client-credentials",
        }

    def test_get_signing_key_from_jwt_by_alg(self):
        token = "eyJhbGciOiJSUzI1NiIsImN0eSI6IkpXVCJ9.eyJpc3MiOiJodHRwczovL2Rldi04N2V2eDlydS5hdXRoMC5jb20vIiwic3ViIjoiYVc0Q2NhNzl4UmVMV1V6MGFFMkg2a0QwTzNjWEJWdENAY2xpZW50cyIsImF1ZCI6Imh0dHBzOi8vZXhwZW5zZXMtYXBpIiwiaWF0IjoxNTcyMDA2OTU0LCJleHAiOjE1NzIwMDY5NjQsImF6cCI6ImFXNENjYTc5eFJlTFdVejBhRTJINmtEME8zY1hCVnRDIiwiZ3R5IjoiY2xpZW50LWNyZWRlbnRpYWxzIn0.gn8boXt4bGSpjuWYijwGE6A0NG1NtRnT07jBw6e2WxBb8KnwxO5BJr-GL0f-UJSRiWDwoSrrwOs5PP0I0XiQPnnmnD4J8OB3z3ETdI3pxa4EsceLaLto0F9SM7JxSHP8NzZJfRwD8GTwgrOs3PrG7nsukvXQYwsRejgYysjsaRCRMa46CfoJGRowYxSuNxtlTMLRlB2q7YNKpxwiCVw1UCrJ_CZybcO3HUFufyuRuWztaI2L8AIueO_oCchhi3X1bNErgzeIza1UsdXrf6Eqf788Easd1YO1RQYSuEejnwdrgh0BERCLMN8kO16vIxYvb2vcM95odRD-ge_lyp8_TA"

        RESPONSE_DATA_NO_KID = RESPONSE_DATA.copy()
        del RESPONSE_DATA_NO_KID["keys"][0]["kid"]

        with mocked_response(RESPONSE_DATA_NO_KID):
            jwks_client = FlaatPyJWKClient(MOCK_URL)
            signing_key = jwks_client.get_signing_key_from_jwt(token)

        data = jwt.decode(
            token,
            signing_key.key,
            algorithms=["RS256"],
            audience="https://expenses-api",
            options={"verify_exp": False},
        )

        assert data == {
            "iss": "https://dev-87evx9ru.auth0.com/",
            "sub": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC@clients",
            "aud": "https://expenses-api",
            "iat": 1572006954,
            "exp": 1572006964,
            "azp": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC",
            "gty": "client-credentials",
        }



from pytest_cases import parametrize
from flaat.test_env import FLAAT_AT


example_paths = {
    "/info",
    "/info_no_strict",
    "/authenticated",
    "/authenticated_callback",
    "/authorized_level",
    "/authorized_claim",
    "/authorized_vo",
}


class Authorized:
    """Request should pass."""

    @parametrize("path", example_paths)
    def case_ValidToken(self, path):
        headers = {"Authorization": f"Bearer {FLAAT_AT}"}
        return path, headers

    @parametrize("path", {"/info_no_strict"})
    def case_FakeToken(self, path):
        headers = {"Authorization": f"Bearer fake_token"}
        return path, headers

    @parametrize("path", {"/info_no_strict"})
    def case_NoBearer(self, path):
        headers = None
        return path, headers


class Unauthorized:
    """Request should not pass."""

    @parametrize("path", example_paths - {"/info_no_strict"})
    def case_FakeToken(self, path):
        headers = {"Authorization": f"Bearer fake_token"}
        return path, headers

    @parametrize("path", example_paths - {"/info_no_strict"})
    def case_NoBearer(self, path):
        headers = None
        return path, headers



import logging
import os
from typing import Callable, List, Optional

import liboidcagent
from attr import dataclass
from dotenv import dotenv_values

from flaat import AuthWorkflow, BaseFlaat
from flaat.exceptions import FlaatException
from flaat.requirements import (
    CheckResult,
    HasClaim,
    get_claim_requirement,
    get_vo_requirement,
)
from flaat.user_infos import UserInfos

logger = logging.getLogger(__name__)

environment = {
    **dotenv_values(".env"),
    **os.environ,
}


def env_var(name, mandatory=True):
    val = environment.get(name, "")
    if mandatory and val == "":  # pragma: no cover
        return ""
        # raise ValueError(f"Set '{name}' in environment or .env file")

    return val


def load_at(short_name: str, mandatory=False, min_valid_period=0, audience=None) -> str:
    try:
        return liboidcagent.get_access_token(
            short_name, min_valid_period=min_valid_period, audience=audience
        )
    except liboidcagent.OidcAgentError as e:  # pragma: no cover
        if mandatory:
            raise FlaatException(
                f"Error acquiring access token for oidc agent account '{short_name}': {e}"
            ) from e
        return ""


OIDC_AGENT_ACCOUNT = env_var("OIDC_AGENT_ACCOUNT")
FLAAT_AT = load_at(OIDC_AGENT_ACCOUNT)

FLAAT_CLAIM_ENTITLEMENT = env_var("FLAAT_CLAIM_ENTITLEMENT")
FLAAT_CLAIM_GROUP = env_var("FLAAT_CLAIM_GROUP")
FLAAT_ISS = env_var("FLAAT_ISS")
FLAAT_TRUSTED_OPS_LIST = [
    FLAAT_ISS,
    "https://accounts.google.com/",  # including google here, because it does not support JWTs
]
FLAAT_CLIENT_ID = environment.get("FLAAT_CLIENT_ID", "")
FLAAT_CLIENT_SECRET = environment.get("FLAAT_CLIENT_SECRET", "")

# optional access token, that is not a JWT
NON_JWT_OIDC_AGENT_ACCOUNT = env_var("NON_JWT_OIDC_AGENT_ACCOUNT", mandatory=False)
NON_JWT_FLAAT_AT = load_at(NON_JWT_OIDC_AGENT_ACCOUNT, mandatory=False)
NON_JWT_FLAAT_ISS = env_var("NON_JWT_FLAAT_ISS", mandatory=False)

# optional oidc agent account from OP that supports setting 'aud' claim
AUD_OIDC_AGENT_ACCOUNT = env_var("AUD_OIDC_AGENT_ACCOUNT", mandatory=False)
AUD_FLAAT_ISS = env_var("AUD_FLAAT_ISS", mandatory=False)


# List to parametrize framework tests
def get_status_kwargs_list():
    return [
        # No access token -> unauthorized
        (401, {"headers": {}}),
        # Invalid access token -> unauthorized
        (401, {"headers": {"Authorization": "Bearer invalid_at"}}),
        # Good access token with the right entitlements
        # If the decorator has a status_code then that one overrides the 200 set here
        # see get_expected_status_code
        (200, {"headers": {"Authorization": f"Bearer {FLAAT_AT}"}}),
    ]


@dataclass
class NamedDecorator:
    name: str
    decorator: Callable
    status_code: Optional[int] = None  # expected status of this decorator

    def get_expected_status_code(self, status_code: int):
        expected = status_code
        if status_code == 200 and self.status_code is not None:
            expected = self.status_code
        return expected


def check_request(user_infos: UserInfos, *args, **kwargs) -> CheckResult:
    _ = user_infos
    _ = args
    _ = kwargs
    return CheckResult(True, "No checks applied")


class User:
    claim_groups: str
    claim_entitlements: str

    groups: List[str]
    entitlements: List[str]

    decorators: List[NamedDecorator]

    def __init__(self, flaat):
        self.flaat: BaseFlaat = flaat
        self.flaat.set_client_id(FLAAT_CLIENT_ID)
        self.flaat.set_client_secret(FLAAT_CLIENT_SECRET)
        self.at = FLAAT_AT
        logger.debug("Fetching user infos for test_env")
        self.user_infos = self.flaat.get_user_infos_from_access_token(self.at)
        if (
            self.user_infos is None or self.user_infos.user_info is None
        ):  # pragma: no cover
            raise FlaatException(
                "Cannot run tests: could not fetch a userinfo with the access token"
            )

        self.claim_groups = FLAAT_CLAIM_GROUP
        self.claim_entitlements = FLAAT_CLAIM_ENTITLEMENT

        self.groups = self.user_infos.user_info.get(self.claim_groups, None)
        if (
            not isinstance(self.groups, list) or len(self.groups) < 2
        ):  # pragma: no cover
            raise FlaatException(
                "FLAAT_CLAIM_GROUP must point to list of at least two groups"
            )

        self.entitlements = self.user_infos.user_info.get(self.claim_entitlements, None)
        if (
            not isinstance(self.entitlements, list) or len(self.entitlements) < 2
        ):  # pragma: no cover
            raise FlaatException(
                "FLAAT_CLAIM_ENTITLEMENT must point to list of at least two entitlements"
            )

    def get_named_decorators(self):
        """construct  decorators for testing"""

        def on_failure(exc, _):
            logger.info("TEST on_failure called")
            raise exc

        decorators = [
            NamedDecorator(
                "inject_user_infos", self.flaat.inject_user_infos(key="test_inject")
            ),
            NamedDecorator(
                "access_level_IDENTIFIED", self.flaat.access_level("IDENTIFIED")
            ),
            NamedDecorator(
                "access_level_IDENTIFIED-on_failure",
                self.flaat.access_level("IDENTIFIED", on_failure=on_failure),
            ),  # with on_failure
            NamedDecorator(
                "requires-GroupAndEntitlement",
                self.flaat.requires(
                    [
                        get_claim_requirement(self.groups, self.claim_groups),
                        get_vo_requirement(self.entitlements, self.claim_entitlements),
                    ],
                ),
            ),  # multiple reqs
            NamedDecorator(
                "requires-forbidden",
                self.flaat.requires(
                    HasClaim("group_that_does_not_exist", self.claim_groups),
                ),
                status_code=403,
            ),  # this must cause forbidden
            NamedDecorator(
                "workflow_all",
                AuthWorkflow(
                    self.flaat,
                    request_requirements=check_request,
                ).decorate_view_func,
            ),
        ]
        return decorators



#!/usr/bin/env python3

import json
import logging
import os
import sys
from typing import Optional

import configargparse
import liboidcagent as agent
from humanfriendly import format_timespan

from flaat import BaseFlaat
from flaat.exceptions import FlaatException
from flaat.user_infos import UserInfos
from flaat.access_tokens import get_access_token_info

logger = logging.getLogger(__name__)

TRUSTED_OP_LIST = [
    "https://b2access.eudat.eu/oauth2/",
    "https://b2access-integration.fz-juelich.de/oauth2",
    "https://unity.helmholtz-data-federation.de/oauth2/",
    "https://login.helmholtz-data-federation.de/oauth2/",
    "https://login-dev.helmholtz.de/oauth2/",
    "https://login.helmholtz.de/oauth2/",
    "https://unity.eudat-aai.fz-juelich.de/oauth2/",
    "https://services.humanbrainproject.eu/oidc/",
    "https://accounts.google.com/",
    "https://aai.egi.eu/oidc/",
    "https://aai.egi.eu/auth/realms/egi",
    "https://aai-demo.egi.eu/auth/realms/egi",
    "https://aai-demo.egi.eu/oidc/",
    "https://aai-dev.egi.eu/oidc/",
    "https://aai-dev.egi.eu/auth/realms/egi",
    "https://login.elixir-czech.org/oidc/",
    "https://iam-test.indigo-datacloud.eu/",
    "https://iam.deep-hybrid-datacloud.eu/",
    "https://iam.extreme-datacloud.eu/",
    "https://oidc.scc.kit.edu/auth/realms/kit/",
    "https://proxy.demo.eduteams.org",
    "https://wlcg.cloud.cnaf.infn.it/",
    "https://proxy.eduteams.org/",
    "https://proxy.eduteams.org/",
    "https://regapp.nfdi-aai.de/oidc/realms/nfdi_demo",
    "https://auth.didmos.nfdi-aai.de",
]


def get_arg_parser():  # pragma: no cover
    path_of_executable = os.path.realpath(sys.argv[0])
    folder_of_executable = os.path.split(path_of_executable)[0]
    full_name_of_executable = os.path.split(path_of_executable)[1]
    name_of_executable = full_name_of_executable.rstrip(".py")

    config_files = [
        f"{os.environ['HOME']}/.config/{name_of_executable}conf",
        f"{folder_of_executable}/{name_of_executable}.conf",
        f"/root/configs/{name_of_executable}.conf",
    ]
    parser = configargparse.ArgumentParser(
        default_config_files=config_files,
        description=name_of_executable,
        ignore_unknown_config_file_keys=True,
    )

    # Arguments below
    parser.add_argument(
        "--my-config",
        "-c",
        is_config_file=True,
        help="config file path",
    )
    parser.add_argument(
        "--client_id",
        default="",
        help="Specify the client_id of an oidc client. This is needed for token introspection.",
    )
    parser.add_argument(
        "--client_secret",
        default="",
        help="Specify the client_secret of an oidc client. This is may be needed for token introspection.",
    )
    parser.add_argument(
        "--oidc-agent-account",
        "-o",
        default="",
        help="Name of oidc-agent account for access token retrieval",
    )
    parser.add_argument(
        "--issuer",
        "-i",
        default="",
        help="Specify issuer (OIDC Provider)",
    )
    parser.add_argument(
        "--audience",
        "--aud",
        default=None,
        help=(
            "Specify an intended audience for the requested access token. "
            "Multiple audiences can be provided as a space separated list. "
            "Only used when token is retrieved via the oidc-agent. "
            "Ignored if OP does not support audience setting."
        ),
    )

    # FLAGS below
    parser.add_argument(
        "--skip_tls_verify",
        default=True,
        action="store_false",
        help="Disable TLS verification",
    )
    parser.add_argument(
        "--skip_jwt_verify",
        default=False,
        action="store_true",
        help="Disable JWT verification",
    )

    parser.add_argument(
        "--accesstoken",
        "-at",
        default=False,
        action="store_true",
        dest="show_access_token",
        help="Show access token info (default)",
    )
    parser.add_argument(
        "--userinfo",
        "-ui",
        default=False,
        action="store_true",
        dest="show_user_info",
        help="Show user info (default)",
    )
    parser.add_argument(
        "--introspection",
        "-in",
        default=False,
        action="store_true",
        dest="show_introspection_info",
        help="Show introspection info (default)",
    )
    parser.add_argument(
        "--all",
        "-a",
        default=True,
        action="store_false",
        dest="show_all",
    )
    parser.add_argument(
        "--quiet",
        "-q",
        default=False,
        action="store_true",
        help="Enable quiet mode. This will only show requested information, no explanatory text",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        dest="verbose",
        help="Enable verbose mode. This will also print debug messages.",
    )
    parser.add_argument(
        "--machine-readable",
        "-m",
        default=False,
        action="store_true",
        dest="machine_readable",
        help="Make stdout machine readable",
    )
    parser.add_argument(
        dest="access_token",
        default=None,
        nargs="*",
        help="An access token (without 'Bearer ')",
    )
    parser.add_argument(
        "--trust-any",
        "--any",
        action="store_true",
        default=True,
        dest="trust_any",
        help="Trust any OP, usefule for displaying information about any access_token",
    )
    return parser


def get_args():
    parser = get_arg_parser()
    args = parser.parse_args()
    # if -in -ui or -at are specified, we set all to false:
    if args.show_user_info or args.show_access_token or args.show_introspection_info:
        args.show_all = False
    return args


def get_flaat(args, trusted_op_list=None):
    flaat = BaseFlaat()

    verbosity = 2  # info
    # quiet has precedence over verbose
    if args.quiet:
        verbosity = 1  # warn
    elif args.verbose:
        verbosity = 3  # debug
    flaat.set_verbosity(verbosity, set_global=True)

    # Log to stderr
    handler = logging.StreamHandler(sys.stderr)
    root = logging.getLogger()
    root.addHandler(handler)

    if trusted_op_list is not None:
        flaat.set_trusted_OP_list(trusted_op_list)
    else:
        flaat.set_trusted_OP_list(TRUSTED_OP_LIST)
    if args.client_id:
        flaat.set_client_id(args.client_id)
    if args.client_secret:
        flaat.set_client_secret(args.client_secret)
    if args.skip_jwt_verify:
        flaat.set_verify_jwt(not args.skip_jwt_verify)
    return flaat


def get_access_token(args) -> Optional[str]:
    access_token = None
    if isinstance(args.access_token, list) and len(args.access_token) > 0:
        # use only the first one for now:
        access_token = args.access_token[0]
        if access_token is not None:
            logger.info("Using access token from commandline")
            return access_token

    # try commandline
    if args.oidc_agent_account != "":
        try:
            access_token = agent.get_access_token(
                args.oidc_agent_account, audience=args.audience
            )
        except agent.OidcAgentError as e:
            raise FlaatException(
                f"Could not use oidc-agent account '{args.oidc_agent_account}': {e}"
            ) from e

        if access_token is not None:
            logger.info(
                "Using access token from oidc-agent (account specified via commandline)"
            )
            return access_token
        raise FlaatException("Access token from oidc-agent is none")

    # try environment  for config
    env_vars_to_try = ["OIDC_AGENT_ACCOUNT"]
    for env_var in env_vars_to_try:
        account_name = os.getenv(env_var)
        if account_name is not None:
            logger.debug("Using agent account '%s'", env_var)
            try:
                access_token = agent.get_access_token(
                    account_name, audience=args.audience
                )
                if access_token is not None:
                    logger.info(
                        "Using access token from oidc-agent (from environment variable '%s')",
                        env_var,
                    )
                    return access_token
            except agent.OidcAgentError as e:
                logger.warning("Could not use oidc-agent: %s", e)

    # try environment for Access Token:
    env_vars_to_try = [
        "ACCESS_TOKEN",
        "OIDC",
        "OS_ACCESS_TOKEN",
        "OIDC_ACCESS_TOKEN",
        "WATTS_TOKEN",
        "WATTSON_TOKEN",
    ]
    for env_var in env_vars_to_try:
        access_token = os.getenv(env_var)
        if access_token is not None:
            logger.info("Using access token from environment variable '%s'", env_var)
            return access_token

    return access_token


class UserInfosPrinter:
    def __init__(self, user_infos: Optional[UserInfos]):
        self.user_infos = user_infos

    def print(self, args):
        if args.machine_readable:
            self.print_machine_readable()
        else:
            self.print_human_readable(args)

    def print_machine_readable(self):
        if self.user_infos is None:
            error = {"error": "No user infos found"}
            print(json.dumps(error))
            sys.exit(2)

        print(self.user_infos.toJSON())

    @staticmethod
    def print_json(content):
        print(
            json.dumps(
                content,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
            )
        )

    def print_human_readable(self, args):
        if self.user_infos is None:
            logger.error("Error: No user infos found")
            sys.exit(2)

        if args.show_access_token or args.show_all:
            if self.user_infos.access_token_info is None:
                logger.warning("Your access token is not a JWT")
            else:
                logger.info("Information stored inside the access token:")
                self.print_json(self.user_infos.access_token_info.__dict__)
            print("")

        if args.show_user_info or args.show_all:
            logger.info("Information retrieved from userinfo endpoint:")
            self.print_json(self.user_infos.user_info)
            print("")

        if args.show_introspection_info or args.show_all:
            if self.user_infos.introspection_info is None:
                if args.client_id != "":
                    logger.warning(
                        """Error retrieving token introspection info
            Submit an issue at https://github.com/indigo-dc/flaat if you feel this is wrong"""
                    )
            else:
                logger.info("Information retrieved from introspection endpoint:")
                self.print_json(self.user_infos.introspection_info)
            print("")

        if self.user_infos.valid_for_secs is not None:
            logger.info(
                "Your token is valid for %s."
                if self.user_infos.valid_for_secs > 0
                else "Your token has EXPIRED for %s!",
                format_timespan(self.user_infos.valid_for_secs),
            )


def main():
    try:
        args = get_args()
        flaat = get_flaat(args)
        access_token = get_access_token(args)
        if access_token is None:
            logger.error("No access token found")
            sys.exit(1)
        if args.trust_any:
            at_info = get_access_token_info(access_token)
            if at_info and hasattr(at_info, "body"):
                flaat.set_trusted_OP_list([at_info.body["iss"]])
        user_infos = flaat.get_user_infos_from_access_token(
            access_token, issuer_hint=args.issuer
        )
        UserInfosPrinter(user_infos).print(args)
    except Exception as e:  # pylint: disable=broad-except
        logger.error("Error: %s", e)
        sys.exit(3)


if __name__ == "__main__":
    main()



import json
import logging
from typing import List

from flaat import BaseFlaat, config

from flask import current_app, request
from flask.wrappers import Response

logger = logging.getLogger(__name__)


# Standard Flask Extension, see:
# https://flask.palletsprojects.com/en/2.1.x/extensiondev/
class Flaat(BaseFlaat):
    def __init__(self, access_levels=config.DEFAULT_ACCESS_LEVELS, app=None):
        self.app = app
        self.access_levels = access_levels
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault("TRUSTED_OP_LIST", [])
        app.config.setdefault("FLAAT_ISS", "")
        app.config.setdefault("FLAAT_CLIENT_ID", "")
        app.config.setdefault("FLAAT_CLIENT_SECRET", "")
        app.config.setdefault("FLAAT_REQUEST_TIMEOUT", 1.2)  # seconds
        app.config.setdefault("FLAAT_VERIFY_TLS", True)
        app.config.setdefault("FLAAT_VERIFY_JWT", True)

    # Now here comes the dirty code to make the think possible, so
    # ATTENTION! This overrides FlaatConfig attr with properties

    @property
    def trusted_op_list(self):
        """Returns the list of Flaat trusted OIDC providers.
        :return: List of strings
        """
        return current_app.config["TRUSTED_OP_LIST"]

    def set_trusted_OP_list(self, trusted_ops: List[str]):
        """Sets a list of OIDC providers that you trust. This means that
        users of these OPs will be able to use your services.
        :param trusted_ops: A list of the issuer URLs that you trust.
            An example issuer is: 'https://iam.deep-hybrid-datacloud.eu/'.
        """
        trusted_ops = list(map(lambda iss: iss.rstrip("/"), trusted_ops))
        current_app.config["TRUSTED_OP_LIST"] = trusted_ops

    @property
    def iss(self):
        """Returns the Flaat configured issuer URL.
        :return: Issuer URL of the pinned issuer
        """
        return current_app.config["FLAAT_ISS"]

    def set_issuer(self, issuer: str):
        """Pins the given issuer. Only users of this issuer will be able
        to use services.
        :param issuer: Issuer URL of the pinned issuer.
        """
        current_app.config["FLAAT_ISS"] = issuer.rstrip("/")

    @property
    def client_id(self):
        """Returns the configured OIDC client id in Flaat.
        :return: OIDC client id
        """
        return current_app.config["FLAAT_CLIENT_ID"]

    def set_client_id(self, client_id=""):
        """Set a client id for token introspection.
        :param client_id: OIDC client id, default is empty string
        """
        # FIXME: consider client_id/client_secret per OP.
        current_app.config["FLAAT_CLIENT_ID"] = client_id

    @property
    def client_secret(self):
        """Returns the configured OIDC client secret in Flaat.
        :return: OIDC client secret
        """
        return current_app.config["FLAAT_CLIENT_SECRET"]

    def set_client_secret(self, client_secret=""):
        """Set a client secret for token introspection
        :param client_secret: OIDC client id, default is empty string
        """
        current_app.config["FLAAT_CLIENT_ID"] = client_secret

    @property
    def request_timeout(self):
        """Returns the timeout for individual requests.
        :return: Request timeout in seconds
        """
        return current_app.config["FLAAT_REQUEST_TIMEOUT"]

    def set_request_timeout(self, timeout: float = 1.2):
        """Set the timeout for individual requests (retrieving issuer configs,
        user infos and introspection infos). Note that the total runtime
        of a decorator could be significantly more, based on your
        `trusted_op_list`.
        :param timeout: Request timeout in seconds
        """
        current_app.config["FLAAT_REQUEST_TIMEOUT"] = timeout

    @property
    def verify_tls(self):
        """Skip TLS certificate verification while processing requests.
        :return: Boolean, indicates if TLS is enabled
        """
        return current_app.config["FLAAT_VERIFY_TLS"]

    def set_verify_tls(self, verify_tls=True):
        """*Only* use for development and debugging. Set to `False` to
        skip TLS certificate verification while processing requests.
        :param verify_tls: Boolean, false disables verification TLS
        """
        current_app.config["FLAAT_VERIFY_TLS"] = verify_tls

    @property
    def verify_jwt(self):
        """Skip JWT verification while processing requests.
        :return: Boolean, indicates if TLS is enabled
        """
        return current_app.config["FLAAT_VERIFY_JWT"]

    def set_verify_jwt(self, verify_jwt=True):
        """Set to `False` to skip JWT verification while processing requests.
        :param verify_jwt: Boolean, false disables JWT verification
        """
        current_app.config["FLAAT_VERIFY_JWT"] = verify_jwt

    # End of dirty code
    # Specific Flask methods to run flaat

    def _get_request(self, *_, **__):
        return request

    def _get_header_from_request(self, _, name) -> str:
        # using flask global "request" here, not an argument
        return request.headers.get(name, "")

    def _make_response(self, data, status_code: int):
        return Response(
            response=json.dumps(data),
            status=status_code,
            mimetype="application/json",
        )



# Flaat example with Flask
import logging
from flaat import AuthWorkflow
from flaat.config import AccessLevel
from flaat.flask import Flaat
from flaat.requirements import CheckResult, HasSubIss, IsTrue
from flaat.requirements import get_claim_requirement
from flaat.requirements import get_vo_requirement

from flask import Blueprint, Flask, abort, current_app
from werkzeug import Response


# ------------------------------------------------------------------
# Basic configuration example ---------------------------------------
logging.basicConfig(level="WARNING")
logging.getLogger("flaat").setLevel("DEBUG")
logging.getLogger("werkzeug").setLevel("DEBUG")

# Standard flask blueprint snippet, source:
# https://flask.palletsprojects.com/en/2.1.x/blueprints
frontend = Blueprint("frontend", "frontend")


# Set a list of access levels to use
def is_admin(user_infos):
    return user_infos.user_info['email'] in current_app.config['ADMIN_EMAILS']


flaat = Flaat([
    AccessLevel("user", HasSubIss()),
    AccessLevel("admin", IsTrue(is_admin)),
])


class Config(object):

    # Defines the list of Flaat trusted OIDC providers
    TRUSTED_OP_LIST = [
        "https://aai-demo.egi.eu/oidc",
        "https://aai-demo.egi.eu/auth/realms/egi",
        "https://aai-dev.egi.eu/oidc",
        "https://aai.egi.eu/oidc/",
        "https://aai.egi.eu/auth/realms/egi",
        "https://accounts.google.com/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://b2access.eudat.eu/oauth2/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.elixir-czech.org/oidc/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://orcid.org/",
        "https://proxy.demo.eduteams.org",
        "https://services.humanbrainproject.eu/oidc/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://wlcg.cloud.cnaf.infn.it/",
        "https://proxy.eduteams.org/",
    ]

    # Additional example configuration:
    ADMIN_EMAILS = ["admin@foo.org", "dev@foo.org"]


# ------------------------------------------------------------------
# Production configuration example ----------------------------------
class ProductionConfig(Config):

    # In production you might want to reduce the number of OP
    TRUSTED_OP_LIST = ["https://aai.egi.eu/oidc/"]
    FLAAT_ISS = "https://aai.egi.eu/oidc/"

    # Define your request timeout for production
    FLAAT_REQUEST_TIMEOUT = 1.2

    # Required for using token introspection endpoint:
    FLAAT_CLIENT_ID = "oidc-agent"
    FLAAT_CLIENT_SECRET = ""


# ------------------------------------------------------------------
# Development configuration example ---------------------------------
class DevelopmentConfig(Config):

    # High timeouts might simplify debugging
    FLAAT_REQUEST_TIMEOUT = 30

    # On development certificate verification might not be needed
    FLAAT_VERIFY_TLS = False
    FLAAT_VERIFY_JWT = False


# ------------------------------------------------------------------
# Testing configuration example -------------------------------------
class TestingConfig(Config):

    # Set TESTING to True to run all Flask plugins on testing mode
    TESTING = True

    # When testing to run requirements as close as possible to production
    FLAAT_REQUEST_TIMEOUT = 1.2


# ------------------------------------------------------------------
# Standard flask Application Factories snippet, source --------------
# https://flask.palletsprojects.com/en/2.1.x/patterns/appfactories
def create_app(config=f"{__name__}.ProductionConfig"):
    app = Flask(__name__)
    app.config.from_object(config)

    # Init application plugins
    flaat.init_app(app)
    # db.init_app(app)
    # mail.init_app(app)

    # Register blueprints
    app.register_blueprint(frontend)
    # app.register_blueprint(admin)
    # app.register_blueprint(other)

    return app


# ------------------------------------------------------------------
# Routes definition -------------------------------------------------
@frontend.route("/", methods=["GET"])
def root():
    text = """This is an example for using flaat with Flask.
    The following endpoints are available:
        /info                       General info about the access_token
        /info_no_strict             General info without token validation
        /authenticated              Requires a valid user
        /authenticated_callback     Requires a valid user, uses a custom callback on error
        /authorized_level           Requires user to fit the specified access level
        /authorized_claim           Requires user to have one of two claims
        /authorized_vo              Requires user to have an entitlement
        /full_custom                Fully custom auth handling
    """
    return Response(text, mimetype="text/plain")


# -------------------------------------------------------------------
# Call with user information ----------------------------------------
@frontend.route("/info", methods=["GET"])
@flaat.inject_user_infos()  # Fail if no valid authentication is provided
def info_strict_mode(user_infos):
    return user_infos.toJSON()


@frontend.route("/info_no_strict", methods=["GET"])
@flaat.inject_user_infos(strict=False)  # Pass with invalid authentication
def info(user_infos=None):
    return user_infos.toJSON() if user_infos else "No userinfo"


# -------------------------------------------------------------------
# Endpoint which requires of an authenticated user ------------------
@frontend.route("/authenticated", methods=["GET"])
@flaat.is_authenticated()
def authenticated():
    return "This worked: there was a valid login"


# -------------------------------------------------------------------
# Instead of giving an error this will return the custom error
# response from `my_on_failure` -------------------------------------
def my_on_failure(exception, user_infos=None):
    text = f"""Custom callback 'my_on_failure' invoked:
        Error Message: {exception}
        User: {user_infos if user_infos else "No Auth"}
    """
    abort(401, description=text)


@frontend.route("/authenticated_callback", methods=["GET"])
@flaat.is_authenticated(on_failure=my_on_failure)
def authenticated_callback():
    return "This worked: there was a valid login"


# -------------------------------------------------------------------
# Endpoint which requires an access level ---------------------------
@frontend.route("/authorized_level", methods=["GET"])
@flaat.access_level("admin")
def authorized_level():
    return "This worked: user has the required rights"


# -------------------------------------------------------------------
# The user needs to satisfy a certain requirement -------------------
email_requirement = get_claim_requirement(
    ["admin@foo.org", "dev@foo.org"],
    claim="email",
    match=1,
)


@frontend.route("/authorized_claim", methods=["GET"])
@flaat.requires(email_requirement)
def authorized_claim():
    return "This worked: User has the claim"


# -------------------------------------------------------------------
# The user needs belong to a certain virtual organization -----------
vo_requirement = get_vo_requirement(
    [
        "urn:mace:egi.eu:group:test:foo",
        "urn:mace:egi.eu:group:test:bar",
    ],
    "mock_entitlements",
    match=2,
)


@frontend.route("/authorized_vo", methods=["GET"])
@flaat.requires(vo_requirement)
def authorized_vo():
    return "This worked: user has the required entitlement"


# -------------------------------------------------------------------
# For maximum customization use AuthWorkflow ------------------------
def my_request_check(user_infos, *args, **kwargs):
    if len(args) != 1:
        return CheckResult(False, "Missing request object")
    return CheckResult(True, "The request is allowed")


def my_process_args(user_infos, *args, **kwargs):
    """We can manipulate the view functions arguments here The user is
    already authenticated at this point, therefore we have `user_infos`,
    therefore we can base our manipulations on the users identity.
    """
    kwargs["email"] = user_infos.get("email", "")
    return (args, kwargs)


custom = AuthWorkflow(
    flaat,  # needs the flaat instance
    user_requirements=get_claim_requirement("bar", "foo"),
    request_requirements=my_request_check,
    process_arguments=my_process_args,
    on_failure=my_on_failure,
    ignore_no_authn=True,  # Don't fail if there is no authentication
)


@frontend.route("/full_custom", methods=["GET"])
@custom.decorate_view_func  # invoke the workflow here
def full_custom(email=""):
    text = f"""This worked: The custom workflow did succeed:
        The users email is: {email}
    """
    return Response(text, mimetype="text/plain")


# -------------------------------------------------------------------
# Main function -----------------------------------------------------
if __name__ == "__main__":
    app = create_app("ProductionConfig")
    app.run(host="0.0.0.0", port=8081)



import pytest

from flaat import BaseFlaat
from flaat.caches import user_infos_cache
from flaat.exceptions import FlaatException, FlaatUnauthenticated
from flaat.test_env import FLAAT_AT, FLAAT_ISS, NON_JWT_FLAAT_AT, NON_JWT_FLAAT_ISS


def get_user_infos(flaat):
    user_infos_cache.clear()
    return flaat.get_user_infos_from_access_token(FLAAT_AT)


def test_success():
    flaat = BaseFlaat()
    flaat.set_trusted_OP_list([FLAAT_ISS])
    info = flaat.get_user_infos_from_access_token(FLAAT_AT)
    assert info is not None


def test_issuer_hint_success():
    flaat = BaseFlaat()
    flaat.set_trusted_OP_list([FLAAT_ISS])
    info = flaat.get_user_infos_from_access_token(FLAAT_AT, issuer_hint=FLAAT_ISS)
    assert info is not None


def test_issuer_hint_fail():
    flaat = BaseFlaat()
    with pytest.raises(FlaatException):
        flaat.get_user_infos_from_access_token(
            FLAAT_AT, issuer_hint="https://invalid.issuer.org"
        )


def test_untrusted():
    flaat = BaseFlaat()
    with pytest.raises(FlaatUnauthenticated):
        get_user_infos(flaat)


def test_invalid_at():
    flaat = BaseFlaat()
    assert flaat.get_user_infos_from_access_token("invalid_at") is None


def test_set_iss():
    flaat = BaseFlaat()

    # correct issuer
    flaat.set_issuer(FLAAT_ISS)
    info = get_user_infos(flaat)
    assert info is not None

    # setting invalid issuer
    flaat.set_issuer("https://another.issuer")
    with pytest.raises(FlaatException):
        get_user_infos(flaat)


def test_non_jwt_at():
    if NON_JWT_FLAAT_AT == "":  # pragma: no cover
        pytest.skip(
            "No non JWT access token (Set NON_JWT_OIDC_AGENT_ACCOUNT to correct shortname)"
        )
    flaat = BaseFlaat()
    flaat.set_trusted_OP_list([NON_JWT_FLAAT_ISS])
    info = flaat.get_user_infos_from_access_token(NON_JWT_FLAAT_AT)
    assert info is not None

    # to test the access_token_issuer_cache we have to bypass the user_infos_cache
    user_infos_cache.clear()

    # this should cause a cache hit in the access_token_issuer_cache
    info = flaat.get_user_infos_from_access_token(NON_JWT_FLAAT_AT)
    assert info is not None



from pytest_cases import parametrize_with_cases
import flaat.flask.flask_test_cases as cases


@parametrize_with_cases("path, headers", cases=cases.Authorized)
def test_authorized(client, path, headers):
    response = client.get(path, headers=headers)
    assert response.status_code == 200


@parametrize_with_cases("path, headers", cases=cases.Unauthorized)
def test_unauthorized(client, path, headers):
    response = client.get(path, headers=headers)
    assert response.status_code == 401



"""
Setup for all tests.

If the tests are run by CI, we mock user infos here.
"""
import logging
from typing import Optional

from pytest import MonkeyPatch

from flaat import BaseFlaat, test_env
from flaat.access_tokens import AccessTokenInfo
from flaat.exceptions import FlaatUnauthenticated
from flaat.test_env import FLAAT_AT, OIDC_AGENT_ACCOUNT
from flaat.user_infos import UserInfos

logger = logging.getLogger(__name__)
logging.getLogger("requests_cache").setLevel(logging.WARN)
logging.getLogger("urllib3").setLevel(logging.WARN)
logging.getLogger("asyncio").setLevel(logging.WARN)


# mock data
_jwt_issuer = "https://mock.issuer.jwt"
_non_jwt_issuer = "https://mock.issuer.non.jwt"
_jwt_user_infos = UserInfos(
    AccessTokenInfo({}, verification=None),
    {
        "iss": _jwt_issuer,
        "sub": "mock_sub",
        "email": "dev@foo.org",
        "mock_groups": ["foo", "bar"],
        "mock_entitlements": [
            "urn:mace:egi.eu:group:test:foo",
            "urn:mace:egi.eu:group:test:bar",
        ],
    },
    None,
)

_non_jwt_user_infos = UserInfos(
    None,
    {
        "iss": _non_jwt_issuer,
        "sub": "non_jwt_mock_sub",
        "mock_groups": ["foo", "bar"],
        "email": "dev@foo.org",
        "mock_entitlements": [
            "urn:mace:egi.eu:group:test:foo",
            "urn:mace:egi.eu:group:test:bar",
        ],
    },
    None,
)


def _mock_get_user_infos_from_access_token(
    self: BaseFlaat, at, issuer_hint=""
) -> Optional[UserInfos]:
    logger.debug("Mock called for access token: %s %s", at, issuer_hint)
    if issuer_hint == "https://invalid.issuer.org":
        raise FlaatUnauthenticated("mock_unauthenticated")
    if at == "invalid_at":
        return None

    info = None
    if at == "mock_jwt_at":
        info = _jwt_user_infos

    if at == "mock_non_jwt_at":
        info = _non_jwt_user_infos

    if info is not None:
        if not self._issuer_is_trusted(info.issuer):
            raise FlaatUnauthenticated(
                f"Issuer {info.issuer} not trusted (trusted: {self.trusted_op_list} {self.iss})"
            )
        return info
    return None


def mock_user_for_ci():
    logger.debug("Monkey patching BaseFlaat as we have no access token")
    mp = MonkeyPatch()
    mp.setattr(
        BaseFlaat,
        "get_user_infos_from_access_token",
        _mock_get_user_infos_from_access_token,
    )
    for (key, value) in [
        ("FLAAT_ISS", _jwt_issuer),
        ("NON_JWT_FLAAT_ISS", _non_jwt_issuer),
        ("FLAAT_TRUSTED_OPS_LIST", [_jwt_issuer, _non_jwt_issuer]),
        ("FLAAT_CLAIM_GROUP", "mock_groups"),
        ("FLAAT_CLAIM_ENTITLEMENT", "mock_entitlements"),
        ("FLAAT_AT", "mock_jwt_at"),
        ("NON_JWT_FLAAT_AT", "mock_non_jwt_at"),
    ]:
        mp.setattr(test_env, key, value)


if OIDC_AGENT_ACCOUNT == "" or FLAAT_AT == "":
    mock_user_for_ci()



# pylint: disable=redefined-outer-name

import logging

import pytest
from aiohttp import web

from flaat.aio import Flaat
from flaat.test_env import FLAAT_TRUSTED_OPS_LIST, User, get_status_kwargs_list

flaat = Flaat()
flaat.set_trusted_OP_list(FLAAT_TRUSTED_OPS_LIST)

DECORATORS = User(flaat).get_named_decorators()

logger = logging.getLogger(__name__)


async def view_func(request, test_inject=None):
    _ = request
    _ = test_inject
    return web.Response(text="Success")


@pytest.fixture
def app():
    """aio web Application for testing"""
    app = web.Application()
    for decorator in DECORATORS:
        decorated = decorator.decorator(view_func)
        app.router.add_get(f"/{decorator.name}", decorated)
    return app


# from: https://docs.aiohttp.org/en/stable/testing.html#pytest-example
@pytest.fixture
def cli(event_loop, aiohttp_client, app):
    return event_loop.run_until_complete(aiohttp_client(app))


@pytest.mark.parametrize("status,kwargs", get_status_kwargs_list())
@pytest.mark.parametrize("decorator", DECORATORS)
async def test_decorator(cli, decorator, status, kwargs):
    logger.debug("Decorator: %s", decorator.name)
    resp = await cli.get(f"/{decorator.name}", **kwargs)
    expected = decorator.get_expected_status_code(status)
    if resp.status != 200:
        err_json = await resp.json()
        logger.debug("Error JSON repsonse: %s", err_json)
    assert resp.status == expected



# This code is distributed under the MIT License

from __future__ import annotations

import json
import logging
import re
from typing import Optional

import requests
from requests.models import HTTPBasicAuth

from flaat.user_infos import UserInfos

logger = logging.getLogger(__name__)

# Defaults for requests
_VERIFY_TLS = True
_TIMEOUT = 1.2  # (seconds)


def is_url(string):
    """Return True if parameter is a URL, otherwise False"""
    regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    if re.match(regex, string):
        return True
    return False


def _make_json_request(
    url, timeout: float, verify_tls: bool, **kwargs
) -> Optional[dict]:
    try:
        if "data" in kwargs:
            resp = requests.post(url, verify=verify_tls, timeout=timeout, **kwargs)
        else:
            resp = requests.get(url, verify=verify_tls, timeout=timeout, **kwargs)

        if resp.status_code != 200:
            logger.debug("Error response: %s %s", resp.status_code, resp.text)
            return None

        resp_json = dict(resp.json())
        return resp_json

    except requests.exceptions.RequestException as e:
        logger.debug("Error making json request to %s: %s", url, e)
        return None


class IssuerConfig:
    issuer_config: dict
    client_id: str
    client_secret: str
    timeout: float
    verify_tls: bool

    def __init__(
        self,
        issuer_config: dict,
        client_id="",
        client_secret="",
        timeout=_TIMEOUT,
        verify_tls=_VERIFY_TLS,
    ):
        self.issuer_config = issuer_config
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout = timeout
        self.verify_tls = verify_tls

    @property
    def issuer(self) -> str:
        return self.issuer_config.get("issuer", "")

    @classmethod
    def _get_from_url(cls, url, timeout, verify_tls) -> Optional[IssuerConfig]:
        config_url = url

        # remove slashes:
        config_url = re.sub("^https?://", "", config_url)
        config_url = config_url.replace("//", "/")
        config_url = "https://" + config_url

        logger.debug("Fetching issuer config from: %s", config_url)
        issuer_config_dict = _make_json_request(config_url, timeout, verify_tls)
        if issuer_config_dict is None:
            return None

        return cls(
            issuer_config=issuer_config_dict, timeout=timeout, verify_tls=verify_tls
        )

    @classmethod
    def get_from_string(
        cls, iss: str, timeout=_TIMEOUT, verify_tls=_VERIFY_TLS
    ) -> Optional[IssuerConfig]:
        """If the string provided is a URL: try several well known endpoints until the ISS config is
        found"""
        if iss is None or not is_url(iss):
            return None

        well_known_path = "/.well-known/openid-configuration"

        if iss.endswith(well_known_path):
            return cls._get_from_url(iss, timeout, verify_tls)

        if iss.endswith(("/oauth2", "/oauth2/")):
            return cls._get_from_url(iss + well_known_path, timeout, verify_tls)

        for url in [
            iss + well_known_path,
            iss + "/oauth2" + well_known_path,
        ]:
            iss_config = cls._get_from_url(url, timeout, verify_tls)
            if iss_config is not None:
                logger.info("Retrieved config for issuer: %s", iss)
                return iss_config

        return None

    def _get_introspected_token_info(self, access_token: str):
        """Query te token introspection endpoint, if there is a client_id and client_secret set"""

        if self.client_id == "" and self.client_secret == "":
            logger.debug(
                "Skipping token introspection because both client_id and client_secret are not configured"
            )
            return None

        introspection_endpoint = self.issuer_config.get("introspection_endpoint", "")
        if introspection_endpoint == "":
            logger.debug(
                "Skipping token introspection because there is no introspection endpoint"
            )
            return None

        post_data = {"token": access_token}
        introspection_info_dict = _make_json_request(
            introspection_endpoint,
            self.timeout,
            self.verify_tls,
            data=post_data,
            auth=HTTPBasicAuth(self.client_id, self.client_secret),
        )
        logger.debug(
            "Got introspection info from %s: %s",
            introspection_endpoint,
            json.dumps(
                introspection_info_dict,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
            ),
        )
        return introspection_info_dict

    def _get_user_info(self, access_token: str) -> Optional[dict]:
        """Query the userinfo endpoint, using the AT as authentication"""

        userinfo_endpoint = self.issuer_config.get("userinfo_endpoint", "")
        if userinfo_endpoint == "":
            return None

        headers = {"Authorization": f"Bearer {access_token}"}
        logger.debug("Trying to get userinfo from %s", userinfo_endpoint)
        user_info_dict = _make_json_request(
            userinfo_endpoint, self.timeout, self.verify_tls, headers=headers
        )
        logger.debug(
            "Got userinfo from %s: %s",
            userinfo_endpoint,
            json.dumps(
                user_info_dict, sort_keys=True, indent=4, separators=(",", ": ")
            ),
        )
        return user_info_dict

    def get_user_infos(
        self,
        access_token,
        access_token_info=None,
    ) -> Optional[UserInfos]:
        user_info = self._get_user_info(access_token)
        if user_info is None:
            return None

        if "iss" not in user_info:
            user_info["iss"] = self.issuer
        introspection_info = self._get_introspected_token_info(access_token)

        return UserInfos(access_token_info, user_info, introspection_info)



import logging

from aiohttp import web
from flaat import AuthWorkflow
from flaat.aio import Flaat
from flaat.exceptions import FlaatException
from flaat.requirements import CheckResult
from flaat.requirements import get_claim_requirement
from flaat.requirements import get_vo_requirement
from flaat.user_infos import UserInfos

# do some log setup so we can see something
logging.basicConfig(level="WARNING")
logging.getLogger("flaat").setLevel("DEBUG")


##########
# aio application
app = web.Application()
routes = web.RouteTableDef()

# Our FLAAT instance
flaat = Flaat()

# Insert OPs that you trust here
flaat.set_trusted_OP_list(
    [
        "https://b2access.eudat.eu/oauth2/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://services.humanbrainproject.eu/oidc/",
        "https://accounts.google.com/",
        "https://login.elixir-czech.org/oidc/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://aai.egi.eu/oidc/",
        "https://aai.egi.eu/auth/realms/egi",
        "https://aai-demo.egi.eu/auth/realms/egi",
        "https://aai-demo.egi.eu/oidc",
        "https://aai-dev.egi.eu/oidc",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://proxy.demo.eduteams.org",
        "https://wlcg.cloud.cnaf.infn.it/",
        "https://proxy.eduteams.org/",
    ]
)

# verbosity:
#     0: Errors
#     1: Warnings
#     2: Infos
#     3: Debug output
flaat.set_verbosity(3)


# # Required for using token introspection endpoint:
# flaat.set_client_id('')
# flaat.set_client_secret('')


# this will customize error responses for the user (used down below)
def my_on_failure(exception: FlaatException, user_infos: UserInfos = None):
    user = "no auth"
    if user_infos is not None:
        user = str(user_infos)
    return web.Response(
        text=f"Custom my_on_failure invoked:\nError Message: {exception}\nUser: {user}"
    )


@routes.get("/")
async def root(request):
    text = """This is an example for useing flaat with AIO. These endpoints are available:
    /info                       General info about the access_token (if provided)
    /authenticated              Requires a valid user
    /authenticated_callback     Requires a valid user, uses a custom callback on error
    /authorized_claim           Requires user to have one of two claims
    /authorized_vo              Requires user to have an entitlement
    /full_custom                Fully custom auth handling
        """
    return web.Response(text=text)


@routes.get("/info")
@flaat.inject_user_infos(
    strict=False,  # we don't fail if there is no user
)
async def info(
    request,
    user_infos: UserInfos = None,  # here is the variable that gets injected, this should have a default value
):
    message = "No userinfo"
    if user_infos is not None:
        message = user_infos.toJSON()

    return web.Response(text=message)


@routes.get("/authenticated")
@flaat.is_authenticated()
async def authenticated_user(request):
    return web.Response(text="This worked: there was a valid login")


@routes.get("/authenticated_callback")
@flaat.is_authenticated(
    on_failure=my_on_failure,  # called if there is no authentication
)
async def valid_user_own_callback(request):
    """instead of giving an error this will return the custom error response from `my_on_failure`"""
    return web.Response(text="This worked: there was a valid login")


@routes.get("/authorized_claim")
@flaat.requires(
    get_claim_requirement(  # the user needs to satisfy this requirement (having one of the email claims)
        ["admin@foo.org", "dev@foo.org"],
        claim="email",
        match=1,
    ),
)
async def authorized_claim(request):
    return web.Response(text="This worked: User has the claim")


@routes.get("/authorized_vo")
@flaat.requires(
    get_vo_requirement(
        [
            "urn:geant:h-df.de:group:m-team:feudal-developers",
            "urn:geant:h-df.de:group:MyExampleColab#unity.helmholtz.de",
        ],
        "eduperson_entitlement",
        match=1,
    )
)
async def authorized_vo(request):
    return web.Response(text="This worked: user has the required entitlement")


# If you need maxmimum customizability, there is AuthWorkflow:

# User requirements
user_reqs = get_claim_requirement("bar", "foo")


def my_request_check(user_infos: UserInfos, *args, **kwargs):
    if len(args) != 1:
        return CheckResult(False, "Missing request object")

    return CheckResult(True, "The request is allowed")


def my_process_args(user_infos: UserInfos, *args, **kwargs):
    """
    We can manipulate the view functions arguments here
    The user is already authenticated at this point, therefore we have `user_infos`,
    therefore we can base our manipulations on the users identity.
    """
    kwargs["email"] = user_infos.get("email", "")
    return (args, kwargs)


custom = AuthWorkflow(
    flaat,  # needs the flaat instance
    user_requirements=user_reqs,
    request_requirements=my_request_check,
    process_arguments=my_process_args,
    on_failure=my_on_failure,
    ignore_no_authn=True,  # Don't fail if there is no authentication
)


@routes.get("/full_custom")
@custom.decorate_view_func  # invoke the workflow here
async def full_custom(request, email=""):
    return web.Response(
        text=f"This worked: The custom workflow did succeed\nThe users email is: {email}"
    )


app.add_routes(routes)

##########
# Main

if __name__ == "__main__":
    web.run_app(app, host="0.0.0.0", port=8080)



"""
Python support for OIDC Access Tokens -- FLAAT.
Use decorators for authorising access to OIDC authenticated REST APIs.
"""
# This code is distributed under the MIT License

from __future__ import annotations

import logging
import os
from asyncio import iscoroutinefunction
from functools import wraps
from typing import Any, Callable, List, NoReturn, Optional, Tuple, Union

from cachetools import cached

from flaat.access_tokens import AccessTokenInfo, get_access_token_info
from flaat.caches import (
    access_token_issuer_cache,
    issuer_config_cache,
    user_infos_cache,
)
from flaat.config import OPS_THAT_SUPPORT_JWT, FlaatConfig
from flaat.exceptions import FlaatException, FlaatForbidden, FlaatUnauthenticated
from flaat.issuers import IssuerConfig
from flaat.requirements import (
    REQUEST_REQUIREMENT,
    REQUIREMENT,
    AllOf,
    HasSubIss,
    Requirement,
)
from flaat.user_infos import UserInfos

logger = logging.getLogger(__name__)

ENV_VAR_AUTHN_OVERRIDE = "DISABLE_AUTHENTICATION_AND_ASSUME_AUTHENTICATED_USER"
ENV_VAR_AUTHZ_OVERRIDE = "DISABLE_AUTHORIZATION_AND_ASSUME_AUTHORIZED_USER"

# MAP_EXCEPTION is the type for self.map_exception
# It map our exceptions to framework specific or custom exceptions
MAP_EXCEPTION = Callable[[FlaatException], NoReturn]

# ON_FAILURE is the type for on_failure functions
# these are called in place of a view function, if the authentication process resulted in an error
# on_failure functions may raise a framework specific exception, or return a valid response for the respective framework.
ON_FAILURE = Callable[[FlaatException, Optional[UserInfos]], Union[Any, NoReturn]]


class BaseFlaat(FlaatConfig):
    """
    Uses OIDC access tokens to provide authentication and authorization for multiple webframe works. This is the base class.
    Use the framework specific classes :class:`flaat.flask.Flaat`, :class:`flaat.aio.Flaat` and :class:`flaat.fastapi.Flaat` directly.

    You usually use a global instance of Flaat, configure it (see :class:`flaat.config.FlaatConfig`) and then access its decorators
    (e.g. :meth:`is_authenticated`, :meth:`requires`, :meth:`inject_object` and :meth:`access_level`).
    """

    @property
    def authentication_disabled(self):
        return "YES" == os.environ.get(ENV_VAR_AUTHN_OVERRIDE, "")

    @property
    def authorization_disabled(self):
        return "YES" == os.environ.get(ENV_VAR_AUTHZ_OVERRIDE, "")

    # SUBCLASS STUBS

    def _get_request(self, *args, **kwargs):  # pragma: no cover
        """overwritten in subclasses"""
        # raise NotImplementedError("implement in subclass")
        _ = args
        _ = kwargs
        return {}

    def _get_header_from_request(self, request, name: str) -> str:  # pragma: no cover
        """overwritten in subclasses
        Extracts a header from a request"""
        _ = request
        _ = name
        return ""

    # pylint: disable=useless-return
    def _make_response(self, data: dict, status_code: int) -> Any:
        """overwritten in subclasses
        returns a json response of the data"""
        _ = data
        _ = status_code
        return None

    # END SUBCLASS STUBS

    def make_response(self, data: dict, status_code: int) -> Any:
        # this call is made to a subclass
        return self._make_response(data, status_code)

    def _get_access_token_from_request(self, request) -> str:
        value = self._get_header_from_request(request, "Authorization")
        if value == "":
            raise FlaatUnauthenticated("No authorization header")

        prefix = "Bearer "
        if not value.startswith(prefix):
            raise FlaatUnauthenticated("No bearer token in authorization header")

        return value.replace(prefix, "")

    def _issuer_is_trusted(self, issuer):
        if self.iss == issuer:  # always trust the issuer pin
            return True
        return issuer.rstrip("/") in self.trusted_op_list

    @cached(cache=issuer_config_cache)
    def _get_issuer_config(self, iss) -> Optional[IssuerConfig]:
        if not self._issuer_is_trusted(iss):
            raise FlaatUnauthenticated(f"Issuer not trusted: {iss}")

        issuer_config = IssuerConfig.get_from_string(
            iss, timeout=self.request_timeout, verify_tls=self.verify_tls
        )
        if issuer_config is None:
            return None

        # FIXME Having per issuer secrets would make more sense
        issuer_config.client_id = self.client_id
        issuer_config.client_secret = self.client_secret

        return issuer_config

    def _find_issuer_config(
        self, access_token, access_token_info: Optional[AccessTokenInfo], issuer_hint=""
    ) -> Optional[IssuerConfig]:
        # Issuer hint provided by user
        if issuer_hint != "":
            iss_config = self._get_issuer_config(issuer_hint)
            if iss_config is None:
                raise FlaatException(
                    f"Unable to retrieve issuer config: Issuer '{issuer_hint}' is probably invalid."
                )
            return iss_config

        # Manually set in the config
        if self.iss != "":
            iss_config = self._get_issuer_config(self.iss)
            if iss_config is None:
                raise FlaatException(
                    f"Unable to retrieve issuer config: Issuer from flaat config '{self.iss}' is probably invalid. Use `set_iss` to set a different issuer."
                )
            return iss_config

        # JWT AT
        if access_token_info is not None:
            logger.debug("Access token is a JWT")
            at_iss = access_token_info.issuer
            if at_iss is not None:
                if not self._issuer_is_trusted(at_iss):
                    raise FlaatUnauthenticated(f"Issuer is not trusted: {at_iss}")

                iss_config = self._get_issuer_config(at_iss)
                if iss_config is None:
                    raise FlaatUnauthenticated(
                        f"Unable to fetch issuer config for: {at_iss}"
                    )

                return iss_config

        # Try AT -> Issuer cache
        if access_token in access_token_issuer_cache:
            logger.debug("Cache hit for access_token")
            issuer = access_token_issuer_cache[access_token]
            iss_config = self._get_issuer_config(issuer)
            if iss_config is not None:
                return iss_config

        return None

    def _get_user_infos_brute_force(self, access_token) -> Optional[UserInfos]:
        logger.info("Issuer could not be determined -> trying all trusted OPs")
        # Nice to have: parallel would speed up things here
        for issuer in self.trusted_op_list:
            # skip OPs that would have provided a JWT
            if issuer in OPS_THAT_SUPPORT_JWT:
                continue

            logger.debug("Trying issuer: %s", issuer)
            issuer_config = self._get_issuer_config(issuer)
            if issuer_config is not None:
                user_infos = issuer_config.get_user_infos(access_token)
                if user_infos is not None:
                    logger.debug(
                        "Found issuer for access token: %s", issuer_config.issuer
                    )
                    access_token_issuer_cache[access_token] = issuer_config.issuer
                    return user_infos

        logger.warning("No trusted OP produced a user info for access token")
        return None

    @cached(cache=user_infos_cache)
    def get_user_infos_from_access_token(
        self, access_token: str, issuer_hint: str = ""
    ) -> Optional[UserInfos]:
        """
        This method is used to retrieve all infos about an user.
        You don't need to call this manually, as the decorators will automatically do it for you.


        :param access_token: The access token of the user. The token must not start with 'Bearer '.
        :return: A :class:`flaat.user_infos.UserInfos` instance with all the infos that could be retrieved.
            If no info could be retrieved, then `None` is returned.
        """
        if access_token == "":
            raise FlaatUnauthenticated("No access token")

        logger.debug("Access token: %s", access_token)
        access_token_info = get_access_token_info(access_token, verify=self.verify_jwt)
        issuer_config = self._find_issuer_config(
            access_token, access_token_info, issuer_hint=issuer_hint
        )
        if issuer_config is not None:
            return issuer_config.get_user_infos(
                access_token, access_token_info=access_token_info
            )

        # Last resort: Try all OPs
        return self._get_user_infos_brute_force(access_token)

    def get_user_infos_from_request(self, request_object) -> Optional[UserInfos]:
        access_token = self._get_access_token_from_request(request_object)
        if access_token == "":
            raise FlaatException("No access token from request")

        user_infos = self.get_user_infos_from_access_token(access_token)
        return user_infos

    def authenticate_user(self, *args, **kwargs) -> Optional[UserInfos]:
        """authenticate user needs the same arguments as the view_func it is called from."""
        request_object = self._get_request(*args, **kwargs)
        user_infos = self.get_user_infos_from_request(request_object)
        return user_infos

    def inject_object(
        self,
        infos_to_object: Optional[Callable[[UserInfos], Any]] = None,
        key="object",
        strict=True,
    ) -> Callable:
        """
        Injects an object into a view function given a method to translate a UserInfos instance into the object.
        This is useful for injecting user model instances.

        :param infos_to_object: A function that translates a :class:`flaat.user_infos.UserInfos` instance to a custom object.
        :param key: The key with which the generated object is injected into the `kwargs` of the view function.
        :param strict: If set to `True` this decorator if fail when there is nothing to inject.
        :return: A decorator for a view function.
        """

        def _add_value_to_kwargs(kwargs: dict, key: str, value) -> dict:
            if key in kwargs:
                logger.warning("Overwriting already existing kwarg: %s", kwargs[key])

            kwargs[key] = value
            return kwargs

        def _infos_to_object(user_infos: UserInfos):
            if infos_to_object is not None:
                return infos_to_object(user_infos)
            return user_infos

        def _inject_object(
            user_infos: UserInfos, *args, **kwargs
        ) -> Tuple[tuple, dict]:
            obj = _infos_to_object(user_infos)
            kwargs = _add_value_to_kwargs(kwargs, key, obj)
            return (args, kwargs)

        return AuthWorkflow(
            self, process_arguments=_inject_object, ignore_no_authn=(not strict)
        ).decorate_view_func

    def inject_user_infos(self, key="user_infos", strict=True) -> Callable:
        """
        A decorator which injects the current users :class:`flaat.user_infos.UserInfos` into
        the view function.

        :param key: The key with which the user info is injected.
        :param strict: If set to `True`, an unauthenticated user will not be able to use the view functions
            and cause an error instead.
        :return: A decorator for a view function.
        """
        return self.inject_object(key=key, strict=strict)

    def requires(
        self,
        requirements: Union[REQUIREMENT, List[REQUIREMENT]],
        on_failure: Optional[ON_FAILURE] = None,
    ) -> Callable:
        """
        This returns a decorator that will make sure, that the user fits the requirements before the view function is called. If the user does not, an exception
            for the respective web framework will be thrown, so the user sees the correct error.

        :param requirements: One :class:`flaat.requirements.Requirement` instance or a list of requirements the user needs to fit to have access to the decorated function.
            If the requirements are wrapped in a callable, it will be lazy evaluated once the view_func is called.
        :param on_failure: Optional function to customize the handling of an error. This function can
            either raise an exception or return a response which should be returned in place of the response
            from the view function.
        :return: A decorator for a view function."""

        return AuthWorkflow(
            self,
            user_requirements=requirements,
            on_failure=on_failure,
        ).decorate_view_func

    def _get_access_level_requirement(self, access_level_name: str) -> Requirement:
        requirement = None
        for level in self.access_levels:
            if level.name == access_level_name:
                requirement = level.requirement
                break

        if requirement is None:
            raise FlaatException(
                f"Access level name '{access_level_name}' not found. Configure in 'access_levels'"
            )

        if callable(requirement):
            # we already get lazy evaluated. So we don't stack levels of lazyness here
            return requirement()

        return requirement

    def access_level(
        self, access_level_name: str, on_failure: Optional[ON_FAILURE] = None
    ) -> Callable:
        """
        :param access_level_name: The name of the access_level that the user needs to use the view function.
        :param on_failure: Can be used to either deliver an error response to the user, or raise a specific exception.
        :return: A decorator, that can be used to decorate a view function.
        """

        return self.requires(
            self._get_access_level_requirement(access_level_name), on_failure=on_failure
        )

    def is_authenticated(self, on_failure: Optional[ON_FAILURE] = None) -> Callable:
        """
        This can be used to make sure that users are identified (as in they have a subject and an issuer).
        If you actually want to access the users infos we recommend using :meth:`inject_user_infos` or
        :meth:`inject_object` instead.

        :param on_failure: Can be used to either deliver an error response to the user, or raise a specific exception.
        :return: A decorator for a view function
        """
        return self.requires(HasSubIss(), on_failure=on_failure)


class AuthWorkflow:
    """
    This class can be used if you need maximum customizability for your decorator.
    It encapsulates the complete workflow of a decorator.

    :param flaat: The flaat instance that is currently in use.
    :param user_requirements: Requirement which the user all needs to match, like with :meth:`requires`.
    :param request_requirements: A callable which determines if a users request is allowed to proceed.
        This function is handy if you want to evaluate the arguments for the view function before against the users permissions.
    :param process_arguments: As with :meth:`inject_object`, this can be used to inject data into the view function.
    :param on_failure: Can be used to either deliver an error response to the user, or raise a specific exception.
    :param ignore_no_authn: If set to `True` a failing authentication of the user will not cause exceptions.
    :returns: A class instance, which is used by decorating view functions with its :meth:`decorate_view_func` method.
    """

    def __init__(
        self,
        flaat: BaseFlaat,
        user_requirements: Optional[Union[REQUIREMENT, List[REQUIREMENT]]] = None,
        request_requirements: Optional[
            Union[REQUEST_REQUIREMENT, List[REQUEST_REQUIREMENT]]
        ] = None,
        process_arguments: Optional[
            Callable[[UserInfos, tuple, dict], Tuple[tuple, dict]]
        ] = None,
        on_failure: Optional[ON_FAILURE] = None,
        ignore_no_authn=False,
    ):
        self.flaat = flaat
        self.user_requirements = user_requirements
        self.request_requirements = request_requirements
        self._process_arguments = process_arguments
        self.on_failure = on_failure
        self.ignore_no_authn = ignore_no_authn

    def authenticate_user(self, *args, **kwargs) -> Optional[UserInfos]:
        return self.flaat.authenticate_user(*args, **kwargs)

    def check_user_authorization(self, user_infos: UserInfos) -> Union[None, NoReturn]:
        if self.user_requirements is None:
            return

        if isinstance(self.user_requirements, list):
            req = AllOf(*self.user_requirements)
        else:
            req = self.user_requirements

        check_result = req.is_satisfied_by(user_infos)
        if check_result.is_satisfied:
            return

        exception = FlaatForbidden(f"User {user_infos} does not meet requirements")
        exception.data = check_result.render()
        raise exception

    def check_request_authorization(
        self, user_infos: UserInfos, *args, **kwargs
    ) -> Union[None, NoReturn]:
        if self.request_requirements is None:
            return

        satisfied = True
        failed_checks = []
        for req in (
            self.request_requirements
            if isinstance(self.request_requirements, list)
            else [self.request_requirements]
        ):
            check_result = req(user_infos, *args, **kwargs)
            if not check_result.is_satisfied:
                failed_checks.append(check_result.render())
                satisfied = False

        if satisfied:
            return

        e = FlaatForbidden(f"Request from user {user_infos} does not meet requirements")
        e.data = failed_checks
        raise e

    def map_exception(self, exception) -> Union[NoReturn, Any]:
        logger.debug("Mapping exception: %s", exception)
        return self.flaat.make_response(exception.render(), exception.status_code)

    def handle_failure(
        self, exception: FlaatException, user_infos: Optional[UserInfos]
    ) -> Union[Any, NoReturn]:
        if self.on_failure is not None:
            logging.debug("Calling on_failure")
            try:
                return self.on_failure(exception, user_infos)
            except FlaatException as e:
                return self.map_exception(e)

        return self.map_exception(exception)

    def handle_no_user_authentication(self, message) -> Union[Any, NoReturn]:
        return self.handle_failure(FlaatUnauthenticated(message), None)

    def handle_no_user_authorization(
        self, message, user_infos: UserInfos
    ) -> Union[Any, NoReturn]:
        return self.handle_failure(FlaatForbidden(message), user_infos)

    def handle_no_request_authorization(
        self, message, user_infos: UserInfos
    ) -> Union[Any, NoReturn]:
        return self.handle_failure(FlaatForbidden(message), user_infos)

    def process_arguments(
        self, user_infos: UserInfos, *args, **kwargs
    ) -> Tuple[tuple, dict]:
        if self._process_arguments is not None:
            return self._process_arguments(user_infos, *args, **kwargs)
        return (args, kwargs)

    def _run_work_flow(self, *args, **kwargs) -> Tuple[tuple, dict]:
        if self.flaat.authentication_disabled:
            logger.info(
                "Authentication and authorization are bypassed: Environment variable is set"
            )
            # No error, but also do nothing else
            return (args, kwargs)

        if self.ignore_no_authn:
            try:
                user_infos = self.authenticate_user(*args, **kwargs)
            except FlaatException:
                user_infos = None
        else:
            user_infos = self.authenticate_user(*args, **kwargs)

        if user_infos is None:
            if self.ignore_no_authn:
                # No error, but also do nothing else
                return (args, kwargs)

            # Fail without user infos
            raise FlaatUnauthenticated("User identity could not be determined")

        if self.flaat.authorization_disabled:
            logger.info("Authorization is bypassed: Environment variable is set")
        else:
            self.check_user_authorization(user_infos)
            self.check_request_authorization(user_infos, *args, **kwargs)

        return self.process_arguments(user_infos, *args, **kwargs)

    def _run_work_flow_safe(
        self, *args, **kwargs
    ) -> Tuple[Tuple[tuple, dict], Optional[Any]]:
        """
        :return: (arguments, error_response)
        """
        try:
            return (self._run_work_flow(*args, **kwargs), None)
        except FlaatException as e:
            return ((args, kwargs), self.handle_failure(e, user_infos=None))

    def decorate_view_func(self, view_func: Callable) -> Callable:
        """
        :param view_func: The view function to decorate.
        :return: The decorated view function.
        """

        @wraps(view_func)
        def wrapper(*args, **kwargs):
            ((args, kwargs), error_response) = self._run_work_flow_safe(*args, **kwargs)
            if error_response is not None:
                return error_response
            return view_func(*args, **kwargs)

        @wraps(view_func)
        async def async_wrapper(*args, **kwargs):
            ((args, kwargs), error_response) = self._run_work_flow_safe(*args, **kwargs)
            if error_response is not None:
                return error_response
            return await view_func(*args, **kwargs)

        if iscoroutinefunction(view_func):
            return async_wrapper
        return wrapper



from cachetools import LRUCache, TTLCache

from flaat.user_infos import UserInfos


class UserInfoCache(LRUCache):
    """This caches user_infos for access tokens for an unspecified time.
    Before yielding UserInfos, the validity of user infos is checked."""

    def __getitem__(self, key):
        def _fail(msg):
            self.__delitem__(key)
            raise KeyError(msg)

        item = super().__getitem__(key)
        if isinstance(item, UserInfos):
            if item.valid_for_secs is None:
                _fail("Cache entry validity can not be determined")
            elif item.valid_for_secs <= 0:  # pragma: no cover
                _fail("Cache entry has expired")
        return item


# cache at most 1024 user infos until they are expired
user_infos_cache = UserInfoCache(maxsize=1024)

# cache issuer configs for an hour
issuer_config_cache = TTLCache(maxsize=128, ttl=3600)

# cache access_token_issuer mappings indefinitely
access_token_issuer_cache = LRUCache(maxsize=1024)



# Standard flask pytest fixture, see:
# https://flask.palletsprojects.com/en/2.1.x/testing/
# pylint: disable=redefined-outer-name
import functools
import copy

from flaat import issuers
from pytest_cases import fixture


@fixture(scope="session", params=["flaat.test_env"])
def configuration(request):
    return request.param


@fixture()
def app(configuration):
    from examples.example_flask import create_app

    app = create_app(configuration)
    app.config["ADMIN_EMAILS"] = ["admin@foo.org", "dev@foo.org"]
    app.config["TESTING"] = True
    # other setup can go here
    yield app
    # clean up / reset resources here


@fixture(scope="function")
def client(app):
    return app.test_client()


@fixture(scope="module", autouse=True)
def patch_user_info():
    original = copy.copy(issuers.IssuerConfig._get_user_info)
    issuers.IssuerConfig._get_user_info = replace_email(
        issuers.IssuerConfig._get_user_info
    )
    issuers.IssuerConfig._get_user_info = add_entitlements(
        issuers.IssuerConfig._get_user_info
    )
    yield
    issuers.IssuerConfig._get_user_info = original


def replace_email(func):
    """Replaces the original email by a mock"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        user_info = func(*args, **kwargs)
        if user_info:
            user_info["email"] = "dev@foo.org"
        return user_info

    return wrapper


def add_entitlements(func):
    """Replaces the original entitlements by a mock"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        user_info = func(*args, **kwargs)
        if user_info:
            user_info["mock_entitlements"] = [
                "urn:mace:egi.eu:group:test:foo",
                "urn:mace:egi.eu:group:test:bar",
            ]
        return user_info

    return wrapper



from dataclasses import dataclass

import pytest
from jwt.api_jwt import decode_complete

from flaat import BaseFlaat, flaat_userinfo
from flaat.flaat_userinfo import (
    TRUSTED_OP_LIST,
    UserInfosPrinter,
    get_access_token,
    get_flaat,
    main,
)
from flaat.test_env import (
    FLAAT_AT,
    FLAAT_CLIENT_ID,
    FLAAT_CLIENT_SECRET,
    FLAAT_ISS,
    OIDC_AGENT_ACCOUNT,
    AUD_OIDC_AGENT_ACCOUNT,
)


@dataclass
class ArgsMock:
    quiet = False
    verbose = True
    issuer = ""
    client_id = FLAAT_CLIENT_ID
    client_secret = FLAAT_CLIENT_SECRET
    access_token = []
    oidc_agent_account = ""
    show_access_token = False
    show_user_info = False
    show_introspection_info = False
    show_all = True
    machine_readable = False
    verify_tls = False
    skip_jwt_verify = False
    audience = None
    trust_any = True

    def parse_args(self):
        return self


@pytest.fixture
def args():
    return ArgsMock()


def test_print_infos(args):
    flaat = BaseFlaat()
    flaat.set_trusted_OP_list([FLAAT_ISS])
    user_infos = flaat.get_user_infos_from_access_token(FLAAT_AT)
    printer = UserInfosPrinter(user_infos)

    printer.print(args)

    args.machine_readable = True
    printer.print(args)


def test_get_flaat():
    trusted_op_list = TRUSTED_OP_LIST[:1]
    assert get_flaat(ArgsMock(), trusted_op_list) is not None


def test_get_at_environment_oidc_agent(args, monkeypatch):
    if OIDC_AGENT_ACCOUNT == "":
        pytest.skip("No oidc agent")
    monkeypatch.setenv("OIDC_AGENT_ACCOUNT", OIDC_AGENT_ACCOUNT)
    assert get_access_token(args) == FLAAT_AT


def test_get_at_environment_access_token(args, monkeypatch):
    monkeypatch.delenv("OIDC_AGENT_ACCOUNT", raising=False)
    monkeypatch.setenv("ACCESS_TOKEN", "bar")
    assert get_access_token(args) == "bar"


def test_get_at_oidc_agent(args):
    if OIDC_AGENT_ACCOUNT == "":
        pytest.skip("No oidc agent")
    args.oidc_agent_account = OIDC_AGENT_ACCOUNT
    assert get_access_token(args) == FLAAT_AT


def test_get_at_access_token(args):
    args.access_token = ["foo"]
    assert get_access_token(args) == "foo"


def test_get_at_oidc_agent_with_aud(args):
    if AUD_OIDC_AGENT_ACCOUNT == "":
        pytest.skip("No oidc agent account for OP with audience")
    args.oidc_agent_account = AUD_OIDC_AGENT_ACCOUNT
    args.audience = "test-audience"

    at = get_access_token(args)
    assert at is not None

    unverified = decode_complete(at, options={"verify_signature": False})
    assert unverified["payload"]["aud"] == "test-audience"


def test_main(args, monkeypatch):
    monkeypatch.delenv("OIDC_AGENT_ACCOUNT", raising=False)
    monkeypatch.setenv("ACCESS_TOKEN", FLAAT_AT)
    monkeypatch.setattr(flaat_userinfo, "get_arg_parser", lambda: args)
    monkeypatch.setattr(flaat_userinfo, "TRUSTED_OP_LIST", [FLAAT_ISS])
    main()



import logging

from fastapi import Request
from fastapi.responses import JSONResponse

from flaat import BaseFlaat
from flaat.exceptions import FlaatException

logger = logging.getLogger(__name__)


class Flaat(BaseFlaat):
    def _get_request(self, *_, **kwargs):
        if "request" not in kwargs:  # pragma: no cover
            raise FlaatException("No request parameter in view function!")

        return kwargs["request"]

    def _get_header_from_request(self, request: Request, name: str) -> str:
        return request.headers.get(name, "")

    def _make_response(self, data, status_code: int):
        return JSONResponse(
            data,
            status_code=status_code,
        )



import logging

from aiohttp.web import Request, json_response

from flaat import BaseFlaat
from flaat.exceptions import FlaatException

logger = logging.getLogger(__name__)


class Flaat(BaseFlaat):
    def _get_request(self, *args, **kwargs):
        for arg in list(args) + list(kwargs.values()):
            if isinstance(arg, Request):
                return arg

        logger.debug("args: %s - kwargs: %s", args, kwargs)
        raise FlaatException(
            f"Need argument 'request' for framework 'aio': Got args={args} kwargs={kwargs}"
        )

    def _get_header_from_request(self, request: Request, name) -> str:
        return request.headers.get(name, "")

    def _make_response(self, data: dict, status_code: int):
        return json_response(
            data=data,
            status=status_code,
        )
